from sqlalchemy.orm import joinedload
import traceback
from chellow.models import Llfc, Session
from chellow.views import chellow_redirect
import chellow.dloads
import csv
from flask import g
import threading
import os
from chellow.utils import hh_format


def content(user):
    sess = f = writer = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'llfcs.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')
        writer.writerow(
            (
                'Chellow Id', 'DNO Code', 'Code', 'Description',
                'Voltage Level', 'Is Substation?', 'Is Import?', 'Valid From',
                'Valid To'))

        for llfc in sess.query(Llfc).order_by(Llfc.id).options(
                joinedload(Llfc.dno), joinedload(Llfc.voltage_level)):
            writer.writerow(
                (
                    str(llfc.id), llfc.dno.dno_code, llfc.code,
                    llfc.description, llfc.voltage_level.code,
                    llfc.is_substation, llfc.is_import,
                    hh_format(llfc.valid_from), hh_format(llfc.valid_to)))
    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    args = (g.user,)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from decimal import Decimal
import csv
from dateutil.relativedelta import relativedelta
from chellow.utils import (
    HH, validate_hh_start, utc_datetime_parse, loads, parse_mpan_core)
from werkzeug.exceptions import BadRequest
from io import StringIO


def parse_date(row, idx, is_finish=False):
    date_str = row[idx].strip()
    try:
        if len(date_str) == 10:
            dt = utc_datetime_parse(date_str, "%Y-%m-%d")
            if is_finish:
                dt = dt + relativedelta(days=1) - HH
        else:
            dt = utc_datetime_parse(date_str, "%Y-%m-%d %H:%M")
        return validate_hh_start(dt)
    except ValueError as e:
        raise BadRequest(
            "Difficulty parsing a date in the row {row} at position "
            "{idx}: {e}".format(row=row, idx=idx, e=e))


class Parser():
    def __init__(self, f):
        self.reader = csv.reader(
            StringIO(str(f.read(), 'utf-8', errors='ignore')),
            skipinitialspace=True)
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        next(iter(self.reader))  # skip title row
        blank_set = set(('',))
        for self.line_number, self.vals in enumerate(self.reader):

            # skip blank lines
            if len(self.vals) == 0 or set(self.vals) == blank_set:
                continue

            bill_type_code = self.vals[0]
            if bill_type_code.startswith('#'):
                continue  # skip comment lines
            account = self.vals[1]
            mpan_core = parse_mpan_core(self.vals[2])
            reference = self.vals[3]
            issue_date = parse_date(self.vals, 4)
            start_date = parse_date(self.vals, 5)
            finish_date = parse_date(self.vals, 6, True)

            kwh = self.to_decimal(7, 'kwh')
            net = self.to_decimal(8, 'net', True)
            vat = self.to_decimal(9, 'vat', True)
            gross = self.to_decimal(10, 'gross', True)

            if len(self.vals) > 11:
                breakdown_str = self.vals[11].strip()
                if len(breakdown_str) == 0:
                    breakdown = {}
                else:
                    try:
                        breakdown = loads(breakdown_str)
                    except SyntaxError as e:
                        raise BadRequest(str(e))
            else:
                raise BadRequest(
                    "For the line, " + str(self.vals) +
                    " there isn't a 'breakdown' field on the end.")

            while self.vals[-1] == '' and len(self.vals) > 12:
                del self.vals[-1]

            reads = []
            for i in range(12, len(self.vals), 11):
                tpr_str = self.vals[i + 4].strip()
                tpr_code = None if len(tpr_str) == 0 else tpr_str.zfill(5)
                reads.append(
                    {
                        'msn': self.vals[i], 'mpan': self.vals[i + 1],
                        'coefficient': self.to_decimal(i + 2, 'coefficient'),
                        'units': self.vals[i + 3], 'tpr_code': tpr_code,
                        'prev_date': parse_date(self.vals, i + 5),
                        'prev_value': Decimal(self.vals[i + 6]),
                        'prev_type_code': self.vals[i + 7],
                        'pres_date': parse_date(self.vals, i + 8),
                        'pres_value': Decimal(self.vals[i + 9]),
                        'pres_type_code': self.vals[i + 10]})

            raw_bill = {
                'bill_type_code': bill_type_code, 'account': account,
                'mpan_core': mpan_core, 'reference': reference,
                'issue_date': issue_date, 'start_date': start_date,
                'finish_date': finish_date, 'kwh': kwh, 'net': net,
                'vat': vat, 'gross': gross, 'breakdown': breakdown,
                'reads': reads
            }
            raw_bills.append(raw_bill)

        return raw_bills

    def to_decimal(self, dec_index, dec_name, is_money=False):
        try:
            dec_str = self.vals[dec_index]
            dec = Decimal(dec_str)
            if is_money:
                dec += Decimal('0.00')
            return dec
        except IndexError:
            raise BadRequest(
                "The field '" + dec_name +
                "' can't be found. It's expected at position " +
                str(dec_index) + " in the list of fields.")
        except ValueError:
            raise BadRequest(
                "The " + dec_name + " field '" + dec_str +
                "' cannot be parsed as a number. The " + dec_name +
                " field is the " + str(dec_index) + " field of " +
                str(self.vals) + ".")



import threading
import csv
import collections
import traceback
import requests
from dateutil.relativedelta import relativedelta
from chellow.models import (
    Contract, RateScript, get_non_core_contract_id, Session)
from chellow.utils import HH, hh_format, utc_datetime_now, utc_datetime_parse
from werkzeug.exceptions import BadRequest
import atexit
from zish import loads
from decimal import Decimal


ELEXON_PORTAL_SCRIPTING_KEY_KEY = 'elexonportal_scripting_key'


def hh(data_source):
    try:
        cache = data_source.caches['rcrc']
    except KeyError:
        cache = data_source.caches['rcrc'] = {}

    for hh in data_source.hh_data:
        try:
            hh['rcrc-rate'] = rcrc = cache[hh['start-date']]
        except KeyError:
            h_start = hh['start-date']
            db_id = get_non_core_contract_id('rcrc')
            rates = data_source.hh_rate(db_id, h_start)['rates']
            try:
                hh['rcrc-rate'] = rcrc = cache[h_start] = float(
                    rates[key_format(h_start)]) / 1000
            except KeyError:
                try:
                    dt = h_start - relativedelta(days=3)
                    hh['rcrc-rate'] = rcrc = cache[h_start] = float(
                        rates[key_format(dt)]) / 1000
                except KeyError:
                    raise BadRequest(
                        "For the RCRC rate script at " + hh_format(dt) +
                        " the rate cannot be found.")

        hh['rcrc-kwh'] = hh['nbp-kwh']
        hh['rcrc-gbp'] = hh['nbp-kwh'] * rcrc


rcrc_importer = None


def key_format(dt):
    return dt.strftime("%d %H:%M Z")


class RcrcImporter(threading.Thread):
    def __init__(self):
        super(RcrcImporter, self).__init__(name="RCRC Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=100)
        self.stopped = threading.Event()
        self.going = threading.Event()

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message)

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                sess = None
                try:
                    sess = Session()
                    self.log("Starting to check RCRCs.")
                    contract = Contract.get_non_core_by_name(sess, 'rcrc')
                    latest_rs = sess.query(RateScript).filter(
                        RateScript.contract_id == contract.id).order_by(
                        RateScript.start_date.desc()).first()
                    latest_rs_id = latest_rs.id
                    latest_rs_start = latest_rs.start_date

                    month_start = latest_rs_start + relativedelta(months=1)
                    month_finish = month_start + relativedelta(months=1) - HH
                    now = utc_datetime_now()
                    if now > month_finish:
                        self.log(
                            "Checking to see if data is available from " +
                            hh_format(month_start) + " to " +
                            hh_format(month_finish) + " on Elexon Portal.")
                        config = Contract.get_non_core_by_name(
                            sess, 'configuration')
                        props = config.make_properties()

                        scripting_key = props.get(
                            ELEXON_PORTAL_SCRIPTING_KEY_KEY)
                        if scripting_key is None:
                            raise BadRequest(
                                "The property " +
                                ELEXON_PORTAL_SCRIPTING_KEY_KEY +
                                " cannot be found in the configuration "
                                "properties.")

                        contract_props = contract.make_properties()
                        url_str = ''.join(
                            (
                                contract_props['url'],
                                'file/download/RCRC_FILE?key=',
                                scripting_key))

                        r = requests.get(url_str, timeout=60)
                        parser = csv.reader(
                            (x.decode() for x in r.iter_lines()),
                            delimiter=',', quotechar='"')
                        next(parser)
                        next(parser)
                        month_rcrcs = {}
                        for values in parser:
                            hh_date = utc_datetime_parse(values[0], "%d/%m/%Y")
                            hh_date += relativedelta(minutes=30*int(values[2]))
                            if month_start <= hh_date <= month_finish:
                                month_rcrcs[key_format(hh_date)] = Decimal(
                                    values[3])
                        if key_format(month_finish) in month_rcrcs:
                            self.log("The whole month's data is there.")
                            script = {'rates': month_rcrcs}
                            contract = Contract.get_non_core_by_name(
                                sess, 'rcrc')
                            rs = RateScript.get_by_id(sess, latest_rs_id)
                            contract.update_rate_script(
                                sess, rs, rs.start_date, month_finish,
                                loads(rs.script))
                            contract.insert_rate_script(
                                sess, month_start, script)
                            sess.commit()
                            self.log(
                                "Added a new rate script starting at " +
                                hh_format(month_start) + ".")
                        else:
                            msg = "There isn't a whole month there yet."
                            if len(month_rcrcs) > 0:
                                msg += " The last date is " + \
                                    sorted(month_rcrcs.keys())[-1]
                            self.log(msg)
                except BaseException:
                    self.log("Outer problem " + traceback.format_exc())
                    sess.rollback()
                finally:
                    self.lock.release()
                    self.log("Finished checking RCRC rates.")
                    if sess is not None:
                        sess.close()

            self.going.wait(30 * 60)
            self.going.clear()


def get_importer():
    return rcrc_importer


def startup():
    global rcrc_importer
    rcrc_importer = RcrcImporter()
    rcrc_importer.start()


@atexit.register
def shutdown():
    if rcrc_importer is not None:
        rcrc_importer.stop()



import traceback
from flask import request, g
from chellow.utils import req_date, req_int, write_row
from chellow.models import Site, Era, SiteEra, Source, Supply, Session
from sqlalchemy import true, or_, null
from sqlalchemy.orm import joinedload
import chellow.dloads
import csv
import sys
import os
import threading
from chellow.views import chellow_redirect


TYPE_ORDER = {'hh': 0, 'amr': 1, 'nhh': 2, 'unmetered': 3}


def content(sess, start_date, finish_date, site_id, user):
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            "site_hh_data.csv", user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')
        write_row(
            writer, "Site Id", "Site Name", "Associated Site Ids", "Sources",
            "Generator Types", "HH Start Clock-Time", "Imported kWh",
            "Displaced kWh", "Exported kWh", "Used kWh", "Parasitic kWh",
            "Generated kWh", "3rd Party Import", "3rd Party Export",
            "Meter Type"
        )

        sites = sess.query(Site).order_by(Site.code)
        if site_id is not None:
            sites = sites.filter(Site.id == site_id)

        for site in sites:
            sources = set()
            generator_types = set()
            assoc = site.find_linked_sites(sess, start_date, finish_date)
            metering_type = ''
            for era in sess.query(Era).join(SiteEra).filter(
                    SiteEra.site == site, SiteEra.is_physical == true(),
                    Era.start_date <= finish_date, or_(
                        Era.finish_date == null(),
                        Era.finish_date >= start_date),
                    Source.code != 'sub').options(
                        joinedload(Era.supply).joinedload(Supply.source),
                        joinedload(Era.supply).joinedload(
                            Supply.generator_type)):
                mtype = era.meter_category
                if metering_type == '' or \
                        TYPE_ORDER[mtype] < TYPE_ORDER[metering_type]:
                    metering_type = mtype
                sources.add(era.supply.source.code)
                generator_type = era.supply.generator_type
                if generator_type is not None:
                    generator_types.add(generator_type.code)

            assoc_str = ','.join(sorted(s.code for s in assoc))
            sources_str = ','.join(sorted(list(sources)))
            generators_str = ','.join(sorted(list(generator_types)))

            for hh in site.hh_data(sess, start_date, finish_date):
                write_row(
                    writer, site.code, site.name, assoc_str, sources_str,
                    generators_str, hh['start_date'], hh['imp_net'],
                    hh['displaced'], hh['exp_net'], hh['used'], hh['exp_gen'],
                    hh['imp_gen'], hh['imp_3p'], hh['exp_3p'], metering_type
                )

    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    start_date = req_date('start')
    finish_date = req_date('finish')
    if 'site_id' in request.values:
        site_id = req_int('site_id')
    else:
        site_id = None

    args = (sess, start_date, finish_date, site_id, g.user)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from decimal import Decimal
import datetime
from chellow.utils import validate_hh_start, parse_mpan_core
from io import StringIO


def parse_date(date_string):
    return validate_hh_start(datetime.datetime.strptime(date_string, "%Y%m%d"))


class Parser():
    def __init__(self, f):
        self.f = StringIO(str(f.read(), 'utf-8', errors='ignore'))
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        for i, line in enumerate(self.f):
            self.line_number = i
            record_type = line[62:66]
            if record_type == "0100":
                account = line[33:41]
                reference = line[41:46]
                start_date = None
                finish_date = None
                net = Decimal(0)
                vat = Decimal(0)
                mpan_core = None
            elif record_type == "1460":
                net += Decimal(line[67:79]) / Decimal(100)
                vat += Decimal(line[85:97]) / Decimal(100)
            elif record_type == "0461":
                mpan_core = parse_mpan_core(line[135:148])
            elif record_type == "0101":
                start_date = parse_date(line[66:74])
                finish_date = parse_date(line[74:82])
            elif record_type == "1500":
                raw_bill = {
                    'bill_type_code': 'N', 'mpan_core': mpan_core,
                    'account': account, 'reference': reference,
                    'issue_date': start_date, 'start_date': start_date,
                    'finish_date': finish_date, 'kwh': Decimal('0.00'),
                    'net': net, 'vat': vat, 'gross': Decimal('0.00'),
                    'breakdown': {}, 'reads': []
                }
                raw_bills.append(raw_bill)

        return raw_bills



from dateutil.relativedelta import relativedelta
from chellow.utils import (
    to_ct, hh_after, c_months_u, get_file_scripts, hh_before, get_file_rates,
    hh_min, to_utc, ct_datetime)
import chellow.computer
import chellow.duos


def hh(ds, rate_period='monthly', est_kw=None):
    for hh in ds.hh_data:
        if hh['ct-is-month-end']:
            _process_hh(ds, rate_period, est_kw, hh)


def _process_hh(ds, rate_period, est_kw, hh):
    month_start, month_finish = next(
        c_months_u(start_year=hh['ct-year'], start_month=hh['ct-month']))

    month_start_ct = to_ct(month_start)
    if month_start_ct.month > 3:
        year = month_start_ct.year
    else:
        year = month_start_ct.year - 1
    financial_year_start = to_utc(ct_datetime(year, 4, 1))
    last_financial_year_start = to_utc(ct_datetime(year - 1, 4, 1))
    financial_year_finish = to_utc(ct_datetime(year + 1, 3, 31, 23, 30))

    est_triad_kws = []
    earliest_triad = None
    for dt in get_file_rates(
            ds.caches, 'triad_dates',
            last_financial_year_start)['triad_dates']:
        triad_hh = None
        earliest_triad = hh_min(earliest_triad, dt)
        try:
            d = next(ds.get_data_sources(dt, dt, financial_year_start))
            chellow.duos.duos_vb(d)
            triad_hh = d.hh_data[0]

            while dt < financial_year_start:
                dt += relativedelta(years=1)

            for d in ds.get_data_sources(dt, dt, financial_year_start):
                chellow.duos.duos_vb(d)
                datum = d.hh_data[0]
                triad_hh['laf'] = datum['laf']
                triad_hh['gsp-kw'] = datum['laf'] * triad_hh['msp-kw']
        except StopIteration:
            triad_hh = {
                'hist-start': dt, 'msp-kw': 0, 'start-date': dt,
                'status': 'before start of MPAN', 'laf': 1, 'gsp-kw': 0
            }
        est_triad_kws.append(triad_hh)

    if ds.site is None:
        era = ds.supply.find_era_at(ds.sess, earliest_triad)
        if era is None or era.get_channel(
                ds.sess, ds.is_import, 'ACTIVE') is None and est_kw is None:
            est_kw = 0.85 * max(datum['msp-kwh'] for datum in ds.hh_data) * 2
        if est_kw is not None:
            for est_datum in est_triad_kws:
                est_datum['msp-kw'] = est_kw
                est_datum['gsp-kw'] = est_datum['msp-kw'] * est_datum['laf']

    gsp_kw = 0
    for i, triad_hh in enumerate(est_triad_kws):
        triad_prefix = 'triad-estimate-' + str(i + 1)
        hh[triad_prefix + '-date'] = triad_hh['hist-start']
        hh[triad_prefix + '-msp-kw'] = triad_hh['msp-kw']
        hh[triad_prefix + '-status'] = triad_hh['status']
        hh[triad_prefix + '-laf'] = triad_hh['laf']
        hh[triad_prefix + '-gsp-kw'] = triad_hh['gsp-kw']
        gsp_kw += triad_hh['gsp-kw']

    hh['triad-estimate-gsp-kw'] = gsp_kw / 3
    polarity = 'import' if ds.llfc.is_import else 'export'
    gsp_group_code = ds.gsp_group_code
    rate = float(
        get_file_rates(
            ds.caches, 'triad_rates',
            month_start)['triad_gbp_per_gsp_kw'][polarity][gsp_group_code])

    hh['triad-estimate-rate'] = rate

    est_triad_gbp = hh['triad-estimate-rate'] * hh['triad-estimate-gsp-kw']

    if rate_period == 'monthly':
        total_intervals = 12

        est_intervals = 1
        hh['triad-estimate-months'] = est_intervals
    else:
        dt = financial_year_start
        total_intervals = 0
        while dt <= financial_year_finish:
            total_intervals += 1
            dt += relativedelta(days=1)

        est_intervals = 0
        for d in ds.get_data_sources(month_start, month_finish):
            for h in d.hh_data:
                if h['ct-decimal-hour'] == 0:
                    est_intervals += 1

        hh['triad-estimate-days'] = est_intervals

    hh['triad-estimate-gbp'] = est_triad_gbp / total_intervals * est_intervals

    if hh['ct-month'] == 3:
        triad_kws = []
        for t_date in get_file_rates(
                ds.caches, 'triad_dates', month_start)['triad_dates']:
            try:
                d = next(ds.get_data_sources(t_date, t_date))
                if ds.supplier_contract is None or \
                        d.supplier_contract == ds.supplier_contract:
                    chellow.duos.duos_vb(d)
                    thh = d.hh_data[0]
                else:
                    thh = {
                        'hist-start': t_date, 'msp-kw': 0,
                        'start-date': t_date, 'status': 'before contract',
                        'laf': 'before contract', 'gsp-kw': 0}
            except StopIteration:
                thh = {
                    'hist-start': t_date, 'msp-kw': 0, 'start-date': t_date,
                    'status': 'before start of supply',
                    'laf': 'before start of supply', 'gsp-kw': 0}

            while t_date < financial_year_start:
                t_date += relativedelta(years=1)

            try:
                d = next(ds.get_data_sources(t_date, t_date))
                if ds.supplier_contract is None or \
                        d.supplier_contract == ds.supplier_contract:
                    chellow.duos.duos_vb(d)
                    thh['laf'] = d.hh_data[0]['laf']
                    thh['gsp-kw'] = thh['laf'] * thh['msp-kw']
            except StopIteration:
                pass

            triad_kws.append(thh)
        gsp_kw = 0

        for i, triad_hh in enumerate(triad_kws):
            pref = 'triad-actual-' + str(i + 1)
            hh[pref + '-date'] = triad_hh['start-date']
            hh[pref + '-msp-kw'] = triad_hh['msp-kw']
            hh[pref + '-status'] = triad_hh['status']
            hh[pref + '-laf'] = triad_hh['laf']
            hh[pref + '-gsp-kw'] = triad_hh['gsp-kw']
            gsp_kw += triad_hh['gsp-kw']

        hh['triad-actual-gsp-kw'] = gsp_kw / 3
        polarity = 'import' if ds.llfc.is_import else 'export'
        gsp_group_code = ds.gsp_group_code
        tot_rate = 0
        for start_date, finish_date, script in get_file_scripts('triad_rates'):
            if start_date <= financial_year_finish and not hh_before(
                    finish_date, financial_year_start):
                start_month = to_ct(start_date).month
                if start_month < 4:
                    start_month += 12

                if finish_date is None:
                    finish_month = 3
                else:
                    finish_month = to_ct(finish_date).month

                if finish_month < 4:
                    finish_month += 12

                rt = get_file_rates(
                    ds.caches, 'triad_rates', start_date
                    )['triad_gbp_per_gsp_kw'][polarity][gsp_group_code]
                tot_rate += (finish_month - start_month + 1) * float(rt)

        rate = tot_rate / 12
        hh['triad-actual-rate'] = rate

        hh['triad-actual-gbp'] = hh['triad-actual-rate'] * \
            hh['triad-actual-gsp-kw']

        era = ds.supply.find_era_at(ds.sess, month_finish)
        est_intervals = 0

        interval = relativedelta(months=1) if \
            rate_period == 'monthly' else relativedelta(days=1)

        dt = month_finish
        while era is not None and dt > financial_year_start:
            est_intervals += 1
            dt -= interval
            if hh_after(dt, era.finish_date):
                era = ds.supply.find_era_at(ds.sess, dt)

        if rate_period == 'monthly':
            hh['triad-all-estimates-months'] = est_intervals
        else:
            hh['triad-all-estimates-days'] = est_intervals
        hh['triad-all-estimates-gbp'] = est_triad_gbp / total_intervals * \
            est_intervals * -1



from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal
from io import StringIO

from chellow.edi_lib import EdiParser, to_decimal
from chellow.models import Era, Session
from chellow.utils import HH, ct_datetime, parse_mpan_core, to_ct, to_utc

from werkzeug.exceptions import BadRequest


read_type_map = {
    '00': 'N', '09': 'N3', '04': 'C', '02': 'E', '11': 'E3', '01': 'EM',
    '03': 'W', '06': 'X', '05': 'CP', '12': 'IF'
}

tmod_map = {
    '0001': '00001',
    '3U93': '00001',
    'URQ1': '00001',
    'Z012': '00001',
    '246N': '00160',
    '0210': '00210',
    'W246': '00276',
    '0043': '00043',
    'DAQ2': '01071',
    'NIQ2': '01072',
    'NQ22': '00210',
    'DQ22': '00043',
    'D7P2': '00043',
    'N7P2': '00210',
    'MDC1': 'kW',
    'URC1': '00001',
    '0206': '00206',
    '0040': '00040',
    '1140': '01140',
    '1139': '01139',
    'A154': '01154',
    'A153': '01153',
    '0039': '00039',
    '0221': '00221',
    '210E': '00210',
    '1150': '01150',
    '1149': '01149',
    '0153': '00153',
    'WQ42': '00184',
    'NQ42': '00210',
    'EQ42': '00187',
    'WDQ3': '01073',
    'EWQ3': '01074',
    'WDQ4': '01075',
    'EWQ4': '01076',
    'NIQ4': '01077',
    'EQ32': '00072',
    'NQ32': '00184',
    '0276': '00276',
    '0277': '00277',
    '0160': '00160',
    '1152': '01152',
    '1151': '01151',
    '1142': '01142',
    '1141': '01141',
    '0249': '00249',
    '0252': '00252',
    '0240': '00240',
    '0244': '00244',
    '0092': '00092',
    '210A': '00210',
    'MAC2': 'kW',
    'DAC2': '00040',
    'NAC2': '00206',
    'NDC2': '00040',
    'NNC2': '00206',
    'MNC2': 'kW',
    '148A': '00148',
    '221B': '00221',
    '080A': '00080',
    'P1M3': '00248',
    'P2M3': '00251',
    'W1M3': '00239',
    'W2M3': '00073',
    'OTM3': '00093',
    'NIM3': '00208',
    'MDM3': 'kW',
    'URQD': '00001',
    'OPE3': '01088',
    'OEC2': '00190',
    '184A': '00184',
    '210J': '00210',
    '187A': '00187',
    '0183': '00183',
    '0071': '00071',
    'DAE7': '00043',
    'NIE7': '00210',
    'OPAC': '00010',
    'MDM1': 'kW',
    'D179': '01139',
    'N179': '01140',
    'O212': '00212',
    '0151': '00151',
    'SG1U': '00001',
    '0184': '00184',
    '0072': '00072',
    'DM22': '00039',
    'NM22': '00221',
    'MM22': 'kW',
    '0044': '00044',
    '0208': '00208',
    'URM1': '00001',
    '1043': '01043',
    '1042': '01042',
    'U393': '00001',
    'D244': '00040',
    'N244': '00206',
    '24D4': '00040',
    '24N4': '00206',
    'OFX2': '00001',
    'D151': '00043',
    'N151': '00210',
    'N184': '01150',
    'D184': '01149',
    'OPF2': '00257',
    'N154': '00221',
    'O2EC': '00190',
    'OFC2': '00159',
    'CMACUF': 'kW',
    '251A': '00251',
    '208D': '00208',
    '073A': '00073',
    '093A': '00093',
    '248A': '00248',
    '239A': '00239',
    'EA1E': 'kVA',
    'MDM2': 'kVA'}


def to_ct_date(component):
    return to_ct(Datetime.strptime(component, "%y%m%d"))


def to_start_date(component):
    return to_utc(to_ct_date(component))


def to_finish_date(component):
    d = to_ct_date(component)
    return to_utc(ct_datetime(d.year, d.month, d.day, 23, 30))


class Parser():
    def __init__(self, f):
        self.parser = EdiParser(
            StringIO(str(f.read(), 'utf-8', errors='ignore')))
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        breakdown = None

        for self.line_number, code in enumerate(self.parser):
            if code == "BCD":
                ivdt = self.parser.elements[0]
                issue_date = to_utc(to_ct_date(ivdt[0]))

                invn = self.parser.elements[2]
                reference = invn[0]
                account = 'SA' + reference[:9]

                btcd = self.parser.elements[5]
                bill_type_code = btcd[0]

                sumo = self.parser.elements[7]
                start_date = to_start_date(sumo[0])
                if to_ct_date(sumo[1]) in (
                        ct_datetime(2020, 4, 1), ct_datetime(2020, 3, 16)):
                    finish_date = to_start_date(sumo[1]) - HH
                else:
                    finish_date = to_finish_date(sumo[1])

            elif code == "MHD":
                type = self.parser.elements[1]
                message_type = type[0]
                if message_type == "UTLBIL":
                    issue_date = None
                    start_date = None
                    finish_date = None
                    account = None
                    reference = None
                    net = Decimal('0.00')
                    vat = Decimal('0.00')
                    gross = Decimal('0.00')
                    kwh = Decimal(0)
                    reads = []
                    bill_type_code = None
                    mpan_core = None
                    breakdown = defaultdict(int, {'raw-lines': []})

            elif code == "CCD":
                ccde = self.parser.elements[1]
                consumption_charge_indicator = ccde[0]

                if consumption_charge_indicator == "1":
                    prdt = self.parser.elements[6]
                    pvdt = self.parser.elements[7]

                    pres_read_date = to_finish_date(prdt[0])
                    prev_read_date = to_finish_date(pvdt[0])

                    tmod = self.parser.elements[3]
                    mtnr = self.parser.elements[4]
                    mloc = self.parser.elements[5]

                    mpan = mloc[0]
                    mpan = mpan[13:15] + ' ' + mpan[15:18] + ' ' + \
                        mpan[18:] + ' ' + mpan[:2] + ' ' + mpan[2:6] + ' ' + \
                        mpan[6:10] + ' ' + mpan[10:13]

                    prrd = self.parser.elements[9]
                    pres_read_type = read_type_map[prrd[1]]
                    prev_read_type = read_type_map[prrd[3]]

                    adjf = self.parser.elements[12]
                    cons = self.parser.elements[13]

                    coefficient = Decimal(adjf[1]) / Decimal(100000)
                    pres_reading_value = Decimal(prrd[0])
                    prev_reading_value = Decimal(prrd[2])
                    msn = mtnr[0]
                    tpr_native = tmod[0]
                    if tpr_native not in tmod_map:
                        raise BadRequest(
                            "The TPR code " + tpr_native +
                            " can't be found in the TPR list for mpan " +
                            mpan + ".")
                    tpr_code = tmod_map[tpr_native]
                    if tpr_code == 'kW':
                        units = 'kW'
                        tpr_code = None
                    elif tpr_code == 'kVA':
                        units = 'kVA'
                        tpr_code = None
                    else:
                        units = 'kWh'
                        kwh += to_decimal(cons) / Decimal('1000')

                    reads.append(
                        {
                            'msn': msn, 'mpan': mpan,
                            'coefficient': coefficient, 'units': units,
                            'tpr_code': tpr_code, 'prev_date': prev_read_date,
                            'prev_value': prev_reading_value,
                            'prev_type_code': prev_read_type,
                            'pres_date': pres_read_date,
                            'pres_value': pres_reading_value,
                            'pres_type_code': pres_read_type})
                elif consumption_charge_indicator == "2":
                    # tcod = self.parser.elements[2]
                    tmod = self.parser.elements[3]
                    mtnr = self.parser.elements[4]
                    mloc = self.parser.elements[5]

                    mpan = mloc[0]
                    mpan = mpan[13:15] + ' ' + mpan[15:18] + ' ' + \
                        mpan[18:] + ' ' + mpan[:2] + ' ' + mpan[2:6] + ' ' + \
                        mpan[6:10] + ' ' + mpan[10:13]

                    prdt = self.parser.elements[6]
                    pvdt = self.parser.elements[7]

                    pres_read_date = to_finish_date(prdt[0])
                    prev_read_date = to_finish_date(pvdt[0])

                    ndrp = self.parser.elements[8]
                    prrd = self.parser.elements[9]
                    pres_read_type = read_type_map[prrd[1]]
                    prev_read_type = read_type_map[prrd[3]]

                    adjf = self.parser.elements[12]
                    cona = self.parser.elements[13]

                    coefficient = Decimal(adjf[1]) / Decimal(100000)
                    pres_reading_value = Decimal(prrd[0])
                    prev_reading_value = Decimal(prrd[2])
                    msn = mtnr[0]
                    tpr_code = tmod[0]
                    if tpr_code not in tmod_map:
                        raise BadRequest(
                            "The TPR code " + tpr_code +
                            " can't be found in the TPR list for mpan " +
                            mpan + ".")
                    tpr = tmod_map[tpr_code]
                    if tpr == 'kW':
                        units = 'kW'
                        tpr = None
                        prefix = 'md-'
                    elif tpr == 'kVA':
                        units = 'kVA'
                        tpr = None
                        prefix = 'md-'
                    else:
                        units = 'kWh'
                        kwh += to_decimal(cona) / Decimal('1000')
                        prefix = tpr + '-'

                    nuct = self.parser.elements[15]
                    breakdown[prefix + 'kwh'] += \
                        to_decimal(nuct) / Decimal('1000')
                    cppu = self.parser.elements[18]
                    rate_key = prefix + 'rate'
                    if rate_key not in breakdown:
                        breakdown[rate_key] = set()
                    breakdown[rate_key].add(
                        to_decimal(cppu) / Decimal('100000'))
                    ctot = self.parser.elements[19]
                    breakdown[prefix + 'gbp'] += \
                        to_decimal(ctot) / Decimal('100')

                    reads.append(
                        {
                            'msn': msn, 'mpan': mpan,
                            'coefficient': coefficient, 'units': units,
                            'tpr_code': tpr, 'prev_date': prev_read_date,
                            'prev_value': prev_reading_value,
                            'prev_type_code': prev_read_type,
                            'pres_date': pres_read_date,
                            'pres_value': pres_reading_value,
                            'pres_type_code': pres_read_type})
                elif consumption_charge_indicator == '3':
                    # tcod = self.parser.elements[2]
                    tmod = self.parser.elements[3]
                    tmod0 = tmod[0]
                    if tmod0 == 'CCL':
                        prefix = kwh_prefix = 'ccl-'
                    elif tmod0 in ['CQFITC', 'CMFITC']:
                        prefix = 'fit-'
                        kwh_prefix = 'fit-msp-'
                    elif tmod0 == 'FITARR':
                        prefix = kwh_prefix = 'fit-reconciliation-'
                    else:
                        tpr_code = tmod0
                        if tpr_code not in tmod_map:
                            raise BadRequest(
                                "The TPR code " + tpr_code +
                                " can't be found in the TPR list for mpan " +
                                mpan + ".")
                        prefix = kwh_prefix = tmod_map[tpr_code] + '-'

                    mtnr = self.parser.elements[4]
                    ndrp = self.parser.elements[8]
                    cona = self.parser.elements[13]
                    nuct = self.parser.elements[15]
                    breakdown[kwh_prefix + 'kwh'] += \
                        to_decimal(nuct) / Decimal('1000')
                    cppu = self.parser.elements[18]

                    rate_key = prefix + 'rate'
                    if rate_key not in breakdown:
                        breakdown[rate_key] = set()
                    breakdown[rate_key].add(
                        to_decimal(cppu) / Decimal('100000'))

                    ctot = self.parser.elements[19]
                    breakdown[prefix + 'gbp'] += \
                        to_decimal(ctot) / Decimal('100')
                elif consumption_charge_indicator == '4':
                    # tcod = self.parser.elements[2]
                    tmod = self.parser.elements[3]
                    tmod0 = tmod[0]

                    mtnr = self.parser.elements[4]
                    ndrp = self.parser.elements[8]
                    if len(ndrp[0]) > 0:
                        breakdown['standing-days'] += \
                            to_decimal(ndrp)
                    cona = self.parser.elements[13]
                    nuct = self.parser.elements[15]
                    cppu = self.parser.elements[18]
                    ctot = self.parser.elements[19]
                    if len(ctot[0]) > 0:
                        breakdown['standing-gbp'] += \
                            to_decimal(ctot) / Decimal('100')
            elif code == "MTR":
                if message_type == "UTLBIL":

                    if mpan_core is None:
                        sess = Session()
                        era = sess.query(Era).filter(
                            Era.imp_supplier_account == account).first()
                        if era is not None:
                            mpan_core = era.imp_mpan_core
                        sess.close()

                    raw_bill = {
                        'bill_type_code': bill_type_code, 'account': account,
                        'mpan_core': mpan_core, 'reference': reference,
                        'issue_date': issue_date, 'start_date': start_date,
                        'finish_date': finish_date, 'kwh': kwh, 'net': net,
                        'vat': vat, 'gross': gross, 'breakdown': breakdown,
                        'reads': reads
                    }
                    raw_bills.append(raw_bill)
                    breakdown = None

            elif code == "MAN":
                madn = self.parser.elements[2]
                '''
                pc_code = madn[3]
                mtc_code = madn[4]
                llfc_code = madn[5]
                '''

                mpan_core = parse_mpan_core(
                    ''.join((madn[0], madn[1], madn[2])))

            elif code == "VAT":
                uvla = self.parser.elements[5]
                net += to_decimal(uvla) / Decimal('100')
                uvtt = self.parser.elements[6]
                vat += to_decimal(uvtt) / Decimal('100')
                ucsi = self.parser.elements[7]
                gross += to_decimal(ucsi) / Decimal('100')

            if breakdown is not None:
                breakdown['raw-lines'].append(self.parser.line)

        return raw_bills



import traceback
import chellow.dloads
import os
import threading
from flask import g, request
from chellow.views import chellow_redirect
from decimal import Decimal, InvalidOperation
import openpyxl
from itertools import chain
from werkzeug.exceptions import BadRequest
from chellow.utils import req_int
from chellow.models import Session, GspGroup
from zish import dumps
from io import BytesIO
import re


def get_value(row, idx):
    val = row[idx].value
    if isinstance(val, str):
        return val.strip()
    else:
        return val


def get_rate(row, idx):
    cell = row[idx]
    if cell.number_format == '#':
        return None

    val = cell.value
    if val is None:
        return None
    elif isinstance(val, str) and len(val) == 0:
        return None
    else:
        try:
            return round(Decimal(val) / Decimal('100'), 5)
        except InvalidOperation as e:
            raise BadRequest(
                "Can't parse the decimal '" + str(val) + "' " + str(e) +
                ".") from e


def get_rag_rate(row, idx):
    val = get_rate(row, idx)
    return get_rag_rate(row, idx - 1) if val is None else val


def get_zero_rate(row, idx):
    val = get_rate(row, idx)
    return Decimal('0.00000') if val is None else val


def get_decimal(row, idx):
    return round(Decimal(get_value(row, idx)), 5)


def to_llfcs(row, idx):
    val = get_value(row, idx)
    llfcs = []
    if isinstance(val, str):
        for v in map(str.strip, val.split(',')):
            if len(v) > 0:
                if '-' in v:
                    start, finish = v.split('-')
                    for i in range(int(start), int(finish) + 1):
                        llfcs.append(str(i))
                else:
                    llfcs.append(v)
    elif isinstance(val, Decimal):
        llfcs.append(str(int(val)))
    elif isinstance(val, float):
        llfcs_str = str(int(val))
        for i in range(0, len(llfcs_str), 3):
            llfcs.append(llfcs_str[i:i+3])
    elif isinstance(val, int):
        val_str = str(val)
        for i in range(0, len(val_str), 3):
            llfcs.append(val_str[i:i+3])
    return [v.zfill(3) for v in llfcs]


def to_pcs(row, idx):
    val = get_value(row, idx)
    llfcs = []
    if isinstance(val, str):
        for v in map(str.strip, val.replace('or', ',').split(',')):
            if len(v) > 0:
                if '-' in v:
                    start, finish = v.split('-')
                    for i in range(int(start), int(finish) + 1):
                        llfcs.append(str(i))
                else:
                    llfcs.append(v)
    elif isinstance(val, Decimal):
        llfcs.append(str(int(val)))
    elif isinstance(val, float):
        llfcs_str = str(int(val))
        for i in range(0, len(llfcs_str), 3):
            llfcs.append(llfcs_str[i:i+3])
    elif isinstance(val, int):
        val_str = str(val)
        for i in range(0, len(val_str), 3):
            llfcs.append(val_str[i:i+3])
    return [v.zfill(2) for v in llfcs]


BAND_WEEKEND = {
    'Monday to Friday': False,
    'Weekends': True,
    'Monday to Friday (Including Bank Holidays) All Year': False,
    'Saturday and Sunday All Year': True}


def str_to_hr(hr_str):
    for sep in (':', '.'):
        if sep in hr_str:
            break
    hours, minutes = map(Decimal, hr_str.strip().split(sep))
    return hours + minutes / Decimal(60)


def val_to_slots(val):
    slots = []
    time_str = None if val is None else val.strip()
    if time_str not in (None, '', '[Start] - [End]'):
        for t_str in time_str.splitlines():
            for sep in ('-', 'to'):
                if sep in t_str:
                    break
            start_str, finish_str = t_str.split(sep)
            slots.append(
                {
                    'start': str_to_hr(start_str),
                    'finish': str_to_hr(finish_str)})
    return slots


def col_match(row, pattern, repeats=1):
    for i, cell in enumerate(row):
        txt = cell.value
        if txt is not None:
            txt_str = ' '.join(str(txt).lower().split())
            if re.search(pattern, txt_str) is not None:
                if repeats == 1:
                    return i
                else:
                    repeats -= 1

    raise BadRequest(
        "Pattern '" + pattern + "' not found in row " +
        ', '.join(str(cell.value) for cell in row))


def tab_lv_hv(sheet, gsp_rates):
    try:
        tariffs = gsp_rates['tariffs']
    except KeyError:
        tariffs = gsp_rates['tariffs'] = {}

    bands = gsp_rates['bands'] = []

    in_tariffs = False
    title_row = None
    for row in sheet.iter_rows():
        val = get_value(row, 0)
        val_0 = None if val is None else ' '.join(val.split())
        if in_tariffs:
            if val_0 is None or len(val_0) == 0:
                in_tariffs = False
            else:
                llfcs_str = ','.join(
                    chain(to_llfcs(row, 1), to_llfcs(row, 10)))
                pcs_str = ','.join(to_pcs(row, 2))
                tariffs[llfcs_str + '_' + pcs_str] = {
                    'description': val_0,
                    'gbp-per-mpan-per-day': get_zero_rate(
                        row, col_match(title_row, 'fixed')),
                    'gbp-per-kva-per-day': get_zero_rate(
                        row, col_match(title_row, '^capacity')),
                    'excess-gbp-per-kva-per-day': get_zero_rate(
                        row, col_match(title_row, 'exce')),
                    'red-gbp-per-kwh': get_rag_rate(
                        row, col_match(title_row, 'red')),
                    'amber-gbp-per-kwh': get_rag_rate(
                        row, col_match(title_row, 'amber')),
                    'green-gbp-per-kwh': get_rag_rate(
                        row, col_match(title_row, 'green')),
                    'gbp-per-kvarh': get_zero_rate(
                        row, col_match(title_row, 'reactive'))}

        elif val_0 == 'Tariff name' or get_value(row, 1) == "Open LLFCs":
            in_tariffs = True
            title_row = row

        if val_0 in BAND_WEEKEND:
            for i, band_name in enumerate(('red', 'amber')):
                for slot in val_to_slots(get_value(row, i+1)):
                    bands.append(
                        {
                            'weekend': BAND_WEEKEND[val_0],
                            'start': slot['start'],
                            'finish': slot['finish'],
                            'band': band_name})


# State for EHV

EHV_BLANK = 0
EHV_BANDS = 1
EHV_TARIFFS = 2


def tab_ehv(sheet, gsp_rates):
    try:
        tariffs = gsp_rates['tariffs']
    except KeyError:
        tariffs = gsp_rates['tariffs'] = {}

    bands = gsp_rates['super_red'] = []

    state = EHV_BLANK
    title_row = None
    for row in sheet.iter_rows():
        val = get_value(row, 0)
        val_0 = None if val is None else ' '.join(str(val).split()).lower()
        if state == EHV_BLANK:
            if val_0 == 'time periods':
                state = EHV_BANDS
                title_row = row
            elif val_0 in (
                    'import unique identifier', 'import llfc'):
                state = EHV_TARIFFS
                title_row = row

        elif state == EHV_TARIFFS:
            for polarity, repeats in (('import', 1), ('export', 2)):
                llfc_val = get_value(
                    row, col_match(title_row, 'llfc', repeats=repeats))
                llfc = None if llfc_val is None else str(llfc_val).strip()
                if llfc not in (None, ''):
                    tariffs[llfc] = {
                        'gbp-per-kwh': get_rate(
                            row, col_match(
                                title_row, polarity + ' super red')),
                        'gbp-per-day': get_zero_rate(
                            row, col_match(title_row, polarity + ' fixed')),
                        'gbp-per-kva-per-day': get_zero_rate(
                            row, col_match(title_row, polarity + ' capacity')),
                        'excess-gbp-per-kva-per-day': get_zero_rate(
                            row, col_match(title_row, polarity + ' exce'))}

        elif state == EHV_BANDS:
            if val_0 in (None, '', 'notes'):
                state = EHV_BLANK
            else:
                period_str = ' '.join(get_value(row, 0).split()).lower()
                periods = []

                if period_str == 'monday to friday nov to feb (excluding ' + \
                        '22nd dec to 4th jan inclusive)':
                    periods.append(
                        {
                            'weekend': False,
                            'start-month': 11,
                            'start-day': 1,
                            'finish-month': 12,
                            'finish-day': 21})
                    periods.append(
                        {
                            'weekend': False,
                            'start-month': 1,
                            'start-day': 5,
                            'finish-month': 2,
                            'finish-day': 'last'})

                elif period_str in (
                        'monday to friday (including bank '
                        'holidays) november to february',
                        'monday to friday nov to feb'):
                    periods.append(
                        {
                            'weekend': False,
                            'start-month': 11,
                            'start-day': 1,
                            'finish-month': 2,
                            'finish-day': 'last'})

                for slot in val_to_slots(get_value(row, 3)):
                    for period in periods:
                        bands.append(
                            {
                                **period,
                                'start_hour': slot['start'],
                                'finish_hour': slot['finish']})


def content(user, file_name, file_like, gsp_group_id):
    f = sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'dno_rates.zish', user)
        f = open(running_name, mode='w')
        gsp_group = GspGroup.get_by_id(sess, gsp_group_id)
        gsp_rates = {}

        if not file_name.endswith('.xlsx'):
            raise BadRequest(
                "The file extension for " + file_name + " isn't recognized.")

        book = openpyxl.load_workbook(
            file_like, data_only=True, read_only=True)

        for sheet in book.worksheets:
            title = sheet.title.strip().lower()
            if title.startswith('annex 1 '):
                tab_lv_hv(sheet, gsp_rates)
            elif title.startswith('annex 2 '):
                tab_ehv(sheet, gsp_rates)

        rs = {
            gsp_group.code: gsp_rates
        }
        f.write(dumps(rs))

    except BaseException:
        f.write(traceback.format_exc())
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_post(session):
    user = g.user
    file_item = request.files["dno_file"]
    gsp_group_id = req_int('gsp_group_id')

    args = user, file_item.filename, BytesIO(file_item.read()), gsp_group_id
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import os
import traceback
from datetime import datetime as Datetime
from dateutil.relativedelta import relativedelta
import pytz
from sqlalchemy import or_, cast, Float, func
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null, true
import time
from chellow.models import (
    Session, Era, Supply, Site, SiteEra, HhDatum, Channel, Bill)
from chellow.utils import HH, hh_format, hh_min, hh_max, req_int
import chellow.computer
import threading
import chellow.dloads
from flask import g, request
from chellow.views import chellow_redirect


def content(year, month, months, supply_id, user):
    tmp_file = sess = None
    try:
        sess = Session()
        supplies = sess.query(Supply).join(Era).distinct().options(
            joinedload(Supply.generator_type))

        if supply_id is None:
            base_name = "supplies_monthly_duration_for_all_supplies_for_" + \
                str(months) + "_to_" + str(year) + "_" + str(month) + ".csv"
        else:
            supply = Supply.get_by_id(sess, supply_id)
            supplies = supplies.filter(Supply.id == supply.id)
            base_name = "supplies_monthly_duration_for_" + str(supply.id) + \
                "_" + str(months) + "_to_" + str(year) + "_" + str(month) + \
                ".csv"
        running_name, finished_name = chellow.dloads.make_names(
            base_name, user)

        tmp_file = open(running_name, "w")

        caches = {}

        start_date = Datetime(year, month, 1, tzinfo=pytz.utc) - \
            relativedelta(months=months-1)

        field_names = (
            'supply-name', 'source-code', 'generator-type', 'month', 'pc-code',
            'msn', 'site-code', 'site-name', 'metering-type',
            'import-mpan-core', 'metered-import-kwh', 'metered-import-net-gbp',
            'metered-import-estimated-kwh', 'billed-import-kwh',
            'billed-import-net-gbp', 'export-mpan-core', 'metered-export-kwh',
            'metered-export-estimated-kwh', 'billed-export-kwh',
            'billed-export-net-gbp', 'problem', 'timestamp')

        tmp_file.write('supply-id,' + ','.join(field_names) + '\n')

        forecast_date = chellow.computer.forecast_date()

        for i in range(months):
            month_start = start_date + relativedelta(months=i)
            month_finish = month_start + relativedelta(months=1) - HH

            for supply in supplies.filter(
                    Era.start_date <= month_finish, or_(
                        Era.finish_date == null(),
                        Era.finish_date >= month_start)):

                generator_type = supply.generator_type
                if generator_type is None:
                    generator_type = ''
                else:
                    generator_type = generator_type.code

                source_code = supply.source.code
                eras = supply.find_eras(sess, month_start, month_finish)
                era = eras[-1]
                metering_type = era.meter_category

                site = sess.query(Site).join(SiteEra).filter(
                    SiteEra.era == era, SiteEra.is_physical == true()).one()

                values = {
                    'supply-name': supply.name, 'source-code': source_code,
                    'generator-type': generator_type,
                    'month': hh_format(month_finish), 'pc-code': era.pc.code,
                    'msn': era.msn, 'site-code': site.code,
                    'site-name': site.name, 'metering-type': metering_type,
                    'problem': ''}

                tmp_file.write(str(supply.id) + ',')

                for is_import, pol_name in [
                        (True, 'import'), (False, 'export')]:
                    if is_import:
                        mpan_core = era.imp_mpan_core
                    else:
                        mpan_core = era.exp_mpan_core

                    values[pol_name + '-mpan-core'] = mpan_core
                    kwh = 0
                    est_kwh = 0

                    if metering_type in ['hh', 'amr']:
                        est_kwh = sess.query(HhDatum.value).join(Channel) \
                            .join(Era).filter(
                                HhDatum.status == 'E',
                                Era.supply_id == supply.id,
                                Channel.channel_type == 'ACTIVE',
                                Channel.imp_related == is_import,
                                HhDatum.start_date >= month_start,
                                HhDatum.start_date <= month_finish).first()
                        if est_kwh is None:
                            est_kwh = 0
                        else:
                            est_kwh = est_kwh[0]

                    if not (is_import and source_code in ('net', 'gen-net')):
                        kwh_sum = sess.query(
                            cast(func.sum(HhDatum.value), Float)
                            ).join(Channel).join(Era).filter(
                            Era.supply_id == supply.id,
                            Channel.channel_type == 'ACTIVE',
                            Channel.imp_related == is_import,
                            HhDatum.start_date >= month_start,
                            HhDatum.start_date <= month_finish).one()[0]
                        if kwh_sum is not None:
                            kwh += kwh_sum

                    values['metered-' + pol_name + '-estimated-kwh'] = est_kwh
                    values['metered-' + pol_name + '-kwh'] = kwh
                    values['metered-' + pol_name + '-net-gbp'] = 0
                    values['billed-' + pol_name + '-kwh'] = 0
                    values['billed-' + pol_name + '-net-gbp'] = 0
                    values['billed-' + pol_name + '-apportioned-kwh'] = 0
                    values['billed-' + pol_name + '-apportioned-net-gbp'] = 0
                    values['billed-' + pol_name + '-raw-kwh'] = 0
                    values['billed-' + pol_name + '-raw-net-gbp'] = 0

                for bill in sess.query(Bill).filter(
                        Bill.supply == supply, Bill.start_date <= month_finish,
                        Bill.finish_date >= month_start):
                    bill_start = bill.start_date
                    bill_finish = bill.finish_date
                    bill_duration = (
                        bill_finish - bill_start).total_seconds() + 30 * 60
                    overlap_duration = (
                        min(bill_finish, month_finish) -
                        max(bill_start, month_start)).total_seconds() + 30 * 60
                    overlap_proportion = float(
                        overlap_duration) / float(bill_duration)
                    values['billed-import-net-gbp'] += \
                        overlap_proportion * float(bill.net)
                    values['billed-import-kwh'] += \
                        overlap_proportion * float(bill.kwh)

                for era in eras:
                    chunk_start = hh_max(era.start_date, month_start)
                    chunk_finish = hh_min(era.finish_date, month_finish)

                    import_mpan_core = era.imp_mpan_core
                    if import_mpan_core is None:
                        continue

                    supplier_contract = era.imp_supplier_contract

                    if source_code in ['net', 'gen-net', '3rd-party']:
                        supply_source = chellow.computer.SupplySource(
                            sess, chunk_start, chunk_finish, forecast_date,
                            era, True, caches)

                        values['metered-import-kwh'] += sum(
                            datum['msp-kwh']
                            for datum in supply_source.hh_data)

                        import_vb_function = supply_source.contract_func(
                            supplier_contract, 'virtual_bill')
                        if import_vb_function is None:
                            values['problem'] += "Can't find the " \
                                "virtual_bill  function in the supplier " \
                                "contract. "
                        else:
                            import_vb_function(supply_source)
                            values['metered-import-net-gbp'] += \
                                supply_source.supplier_bill['net-gbp']

                        supply_source.contract_func(
                            era.dc_contract, 'virtual_bill')(supply_source)
                        values['metered-import-net-gbp'] += \
                            supply_source.dc_bill['net-gbp']

                        mop_func = supply_source.contract_func(
                            era.mop_contract, 'virtual_bill')
                        if mop_func is None:
                            values['problem'] += " MOP virtual_bill " \
                                "function can't be found."
                        else:
                            mop_func(supply_source)
                            mop_bill = supply_source.mop_bill
                            values['metered-import-net-gbp'] += \
                                mop_bill['net-gbp']
                            if len(mop_bill['problem']) > 0:
                                values['problem'] += \
                                    " MOP virtual bill problem: " + \
                                    mop_bill['problem']

                values['timestamp'] = int(time.time() * 1000)
                tmp_file.write(
                    ','.join(
                        '"' + str(values[name]) +
                        '"' for name in field_names) + '\n')
    except BaseException:
        tmp_file.write(traceback.format_exc())
    finally:
        if sess is not None:
            sess.close()
        tmp_file.close()
        os.rename(running_name, finished_name)


def do_get(sess):
    year = req_int('end_year')
    month = req_int('end_month')
    months = req_int('months')
    supply_id = req_int('supply_id') if 'supply_id' in request.values else None
    user = g.user
    threading.Thread(
        target=content, args=(year, month, months, supply_id, user)).start()
    return chellow_redirect("/downloads", 303)



import csv
from decimal import Decimal
from datetime import datetime as Datetime
import itertools
from chellow.utils import (
    parse_channel_type, parse_mpan_core, validate_hh_start, to_utc)
from werkzeug.exceptions import BadRequest


def create_parser(reader, mpan_map):
    return HhParserCsvSimple(reader, mpan_map)


class HhParserCsvSimple():
    def __init__(self, reader, mpan_map):
        self.shredder = zip(itertools.count(1), csv.reader(reader))
        next(self.shredder)  # skip the title line
        self.values = None

    def get_field(self, index, name):
        if len(self.values) > index:
            return self.values[index].strip()
        else:
            raise BadRequest(
                "Can't find field " + index + ", " + name + ".")

    def __iter__(self):
        return self

    def __next__(self):
        try:
            self.line_number, self.values = next(self.shredder)
            mpan_core_str = self.get_field(0, "MPAN Core")
            datum = {'mpan_core': parse_mpan_core(mpan_core_str)}
            channel_type_str = self.get_field(1, "Channel Type")
            datum['channel_type'] = parse_channel_type(channel_type_str)

            start_date_str = self.get_field(2, "Start Date")
            datum['start_date'] = validate_hh_start(
                to_utc(Datetime.strptime(start_date_str, "%Y-%m-%d %H:%M")))

            value_str = self.get_field(3, "Value")
            datum['value'] = Decimal(value_str)

            status = self.get_field(4, "Status")
            if len(status) != 1:
                raise BadRequest(
                    "The status character must be one character in length.")
            datum['status'] = status
            return datum
        except BadRequest as e:
            e.description = ''.join(
                "Problem at line number: ", str(self.line_number), ": ",
                str(self.values), ": ", e.description)
            raise e

    def close(self):
        self.shredder.close()



import traceback
from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true
from chellow.models import Supply, Era, Site, SiteEra, Session
from chellow.utils import hh_format, hh_range, req_int, req_date, csv_make_val
import chellow.computer
import csv
from flask import g
import threading
from chellow.views import chellow_redirect
import sys
import os


def content(supply_id, start_date, finish_date, user):
    caches = {}
    try:
        sess = Session()
        supply = Supply.get_by_id(sess, supply_id)

        forecast_date = chellow.computer.forecast_date()

        prev_titles = None
        running_name, finished_name = chellow.dloads.make_names(
            'supply_virtual_bills_hh_' + str(supply_id) + '.csv', user)
        f = open(running_name, mode='w', newline='')
        w = csv.writer(f, lineterminator='\n')

        for hh_start in hh_range(caches, start_date, finish_date):
            era = sess.query(Era).filter(
                Era.supply == supply, Era.start_date <= hh_start,
                or_(
                    Era.finish_date == null(),
                    Era.finish_date >= hh_start)).one()

            site = sess.query(Site).join(SiteEra).filter(
                SiteEra.era == era, SiteEra.is_physical == true()).one()

            ds = chellow.computer.SupplySource(
                sess, hh_start, hh_start, forecast_date, era, True, caches)

            titles = [
                'MPAN Core', 'Site Code', 'Site Name', 'Account', 'HH Start',
                '']

            output_line = [
                ds.mpan_core, site.code, site.name, ds.supplier_account,
                hh_format(ds.start_date), '']

            mop_titles = ds.contract_func(
                era.mop_contract, 'virtual_bill_titles')()
            titles.extend(['mop-' + t for t in mop_titles])

            ds.contract_func(era.mop_contract, 'virtual_bill')(ds)
            bill = ds.mop_bill
            for title in mop_titles:
                output_line.append(csv_make_val(bill.get(title, '')))
                if title in bill:
                    del bill[title]
            for k in sorted(bill.keys()):
                output_line.extend([k, csv_make_val(bill[k])])

            output_line.append('')
            dc_titles = ds.contract_func(
                era.dc_contract, 'virtual_bill_titles')()
            titles.append('')
            titles.extend(['dc-' + t for t in dc_titles])

            ds.contract_func(era.dc_contract, 'virtual_bill')(ds)
            bill = ds.dc_bill
            for title in dc_titles:
                output_line.append(csv_make_val(bill.get(title, '')))
                if title in bill:
                    del bill[title]

            for k in sorted(bill.keys()):
                output_line.extend([k, csv_make_val(bill[k])])

            if era.imp_supplier_contract is not None:
                contract = era.imp_supplier_contract
                output_line.append('')
                supplier_titles = ds.contract_func(
                    contract, 'virtual_bill_titles')()
                titles.append('')
                titles.extend(['imp-supplier-' + t for t in supplier_titles])

                ds.contract_func(contract, 'virtual_bill')(ds)
                bill = ds.supplier_bill
                for title in supplier_titles:
                    output_line.append(csv_make_val(bill.get(title, '')))
                    if title in bill:
                        del bill[title]

                for k in sorted(bill.keys()):
                    output_line.extend([k, csv_make_val(bill[k])])

            if era.exp_supplier_contract is not None:
                contract = era.exp_supplier_contract
                ds = chellow.computer.SupplySource(
                    sess, hh_start, hh_start, forecast_date, era, False,
                    caches)
                output_line.append('')
                supplier_titles = ds.contract_func(
                    contract, 'virtual_bill_titles')()
                titles.append('')
                titles.extend(['exp-supplier-' + t for t in supplier_titles])

                ds.contract_func(contract, 'virtual_bill')(ds)
                bill = ds.supplier_bill
                for title in supplier_titles:
                    output_line.append(csv_make_val(bill.get(title, '')))
                    if title in bill:
                        del bill[title]

                for k in sorted(bill.keys()):
                    output_line.extend([k, csv_make_val(bill[k])])

            if titles != prev_titles:
                prev_titles = titles
                w.writerow(titles)
            w.writerow(output_line)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        w.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    supply_id = req_int('supply_id')
    start_date = req_date('start')
    finish_date = req_date('finish')

    args = supply_id, start_date, finish_date, g.user
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from collections import defaultdict
from decimal import Decimal

from chellow.edi_lib import parse_edi, to_date, to_decimal
from chellow.utils import HH

from dateutil.relativedelta import relativedelta


READ_TYPE_MAP = {
    '00': 'A',
    '01': 'E'
}

TCOD_MAP = {
    'Unidentified Gas': {
        'PPK': 'ug'
    },
    'Commodity': {
        'PPK': 'commodity'
    },
    'Transportation': {
        'PPD': 'transportation_fixed',
        'PPK': 'transportation_variable'
    },
    'Gas Flexi': {
        'PPK': 'commodity'
    },
    'Flex - Gas Flexi (New)': {
        'PPK': 'commodity'
    },
    'Meter Reading': {
        'PPD': 'meter_read'
    },
    'Meter Reading Credit Oct 19': {
        'FIX': 'meter_read'
    },
    'Meter Rental': {
        'PPD': 'metering'
    },
    'CCL': {
        'PPK': 'ccl'
    },
    'Consumption Based Administration': {
        'PPK': 'admin_variable'
    },
    'Swing': {
        'PPK': 'swing'
    }
}

UNIT_MAP = {
    'M3': 'M3',
    'HH': 'HCUF'
}


class Parser():
    def __init__(self, file_bytes):
        self.edi_str = str(file_bytes, 'utf-8', errors='ignore')
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        message_type = breakdown = raw_lines = None
        for self.line_number, line, code, elements in parse_edi(self.edi_str):
            if code == "ADJ":
                adjf = elements['ADJF']
                if adjf[0] == 'CV':
                    cv = Decimal(adjf[1]) / Decimal(100000)

            if code == "BCD":
                ivdt = elements['IVDT']
                issue_date = to_date(ivdt[0])

                invn = elements['INVN']
                reference = invn[0]

                btcd = elements['BTCD']
                bill_type_code = btcd[0]

                sumo = elements['SUMO']
                start_date = to_date(sumo[0])
                finish_date = to_date(sumo[1]) + relativedelta(days=1) - HH

            elif code == "MHD":
                typ = elements['TYPE']
                message_type = typ[0]
                if message_type == "UTLBIL":
                    issue_date = None
                    start_date = None
                    finish_date = None
                    reference = None
                    net = Decimal('0.00')
                    vat = Decimal('0.00')
                    gross = Decimal('0.00')
                    kwh = Decimal(0)
                    reads = []
                    bill_type_code = None
                    mprn = None
                    raw_lines = []
                    breakdown = defaultdict(
                        int, {
                            'units_consumed': Decimal(0)
                        }
                    )
                    cv = None
            elif code == "CCD":
                ccde = elements['CCDE']
                consumption_charge_indicator = ccde[0]

                if consumption_charge_indicator == "1":

                    mtnr = elements['MTNR']
                    msn = mtnr[0]

                    mloc = elements['MLOC']

                    # Bug in EDI where MPRN missing in second CCD 1
                    if mprn is None:
                        mprn = mloc[0]

                    prdt = elements['PRDT']
                    pvdt = elements['PVDT']

                    pres_read_date = to_date(prdt[0])
                    prev_read_date = to_date(pvdt[0])

                    prrd = elements['PRRD']
                    pres_read_value = Decimal(prrd[0])
                    pres_read_type = READ_TYPE_MAP[prrd[1]]
                    prev_read_value = Decimal(prrd[2])
                    prev_read_type = READ_TYPE_MAP[prrd[3]]

                    conb = elements['CONB']
                    unit = UNIT_MAP[conb[1]]
                    breakdown['units_consumed'] += \
                        to_decimal(conb) / Decimal('1000')

                    adjf = elements['ADJF']
                    correction_factor = Decimal(adjf[1]) / Decimal(100000)

                    nuct = elements['NUCT']

                    kwh += to_decimal(nuct) / Decimal('1000')

                    reads.append(
                        {
                            'msn': msn,
                            'unit': unit,
                            'correction_factor': correction_factor,
                            'prev_date': prev_read_date,
                            'prev_value': prev_read_value,
                            'prev_type_code': prev_read_type,
                            'pres_date': pres_read_date,
                            'pres_value': pres_read_value,
                            'pres_type_code': pres_read_type
                        }
                    )

                elif consumption_charge_indicator == "2":
                    ccde_supplier_code = ccde[2]
                    tcod = elements['TCOD']

                    tpref = TCOD_MAP[tcod[1]][ccde_supplier_code]

                    mtnr = elements['MTNR']
                    mloc = elements['MLOC']
                    prdt = elements['PRDT']
                    pvdt = elements['PVDT']
                    prrd = elements['PRRD']
                    conb = elements['CONB']
                    adjf = elements['ADJF']

                    bpri = elements['BPRI']
                    rate_key = tpref + '_rate'
                    if rate_key not in breakdown:
                        breakdown[rate_key] = set()
                    rate = Decimal(bpri[0]) / Decimal('10000000')
                    breakdown[rate_key].add(rate)

                    nuct = elements['NUCT']

                    try:
                        ctot = elements['CTOT']
                        breakdown[tpref + '_gbp'] += \
                            to_decimal(ctot) / Decimal('100')

                        if ccde_supplier_code == 'PPK':
                            key = tpref + '_kwh'
                        elif ccde_supplier_code == 'PPD':
                            key = tpref + '_days'

                        breakdown[key] += to_decimal(nuct) / Decimal('1000')
                    except KeyError:
                        pass

                elif consumption_charge_indicator == '3':
                    ccde_supplier_code = ccde[2]
                    tcod = elements['TCOD']

                    tpref = TCOD_MAP[tcod[1]][ccde_supplier_code]

                    mtnr = elements['MTNR']
                    mloc = elements['MLOC']
                    prdt = elements['PRDT']
                    pvdt = elements['PVDT']
                    prrd = elements['PRRD']
                    conb = elements['CONB']
                    adjf = elements['ADJF']
                    bpri = elements['BPRI']
                    rate_key = tpref + '_rate'
                    if rate_key not in breakdown:
                        breakdown[rate_key] = set()
                    rate = Decimal(bpri[0]) / Decimal('10000000')
                    breakdown[rate_key].add(rate)

                    nuct = elements['NUCT']

                    try:
                        ctot = elements['CTOT']
                        breakdown[tpref + '_gbp'] += \
                            to_decimal(ctot) / Decimal('100')

                        if ccde_supplier_code == 'PPK':
                            key = tpref + '_kwh'
                        elif ccde_supplier_code == 'PPD':
                            key = tpref + '_days'

                        breakdown[key] += to_decimal(nuct) / Decimal('1000')
                    except KeyError:
                        pass

                elif consumption_charge_indicator == "4":
                    ccde_supplier_code = ccde[2]
                    tcod = elements['TCOD']

                    tpref = TCOD_MAP[tcod[1]][ccde_supplier_code]

                    mtnr = elements['MTNR']
                    mloc = elements['MLOC']
                    prdt = elements['PRDT']
                    pvdt = elements['PVDT']
                    prrd = elements['PRRD']
                    conb = elements['CONB']
                    adjf = elements['ADJF']
                    bpri = elements['BPRI']
                    rate_key = tpref + '_rate'
                    if rate_key not in breakdown:
                        breakdown[rate_key] = set()
                    rate = Decimal(bpri[0]) / Decimal('10000000')
                    breakdown[rate_key].add(rate)

                    nuct = elements['NUCT']

                    try:
                        ctot = elements['CTOT']
                        breakdown[tpref + '_gbp'] += \
                            to_decimal(ctot) / Decimal('100')

                        if ccde_supplier_code == 'PPK':
                            key = tpref + '_kwh'
                        elif ccde_supplier_code == 'PPD':
                            key = tpref + '_days'

                        breakdown[key] += to_decimal(nuct) / Decimal('1000')
                    except KeyError:
                        pass

            elif code == "MTR":
                if message_type == 'UTLBIL':
                    for k, v in tuple(breakdown.items()):
                        if isinstance(v, set):
                            breakdown[k] = sorted(v)

                    for read in reads:
                        read['calorific_value'] = cv

                    raw_bills.append(
                        {
                            'raw_lines': '\n'.join(raw_lines),
                            'mprn': mprn,
                            'reference': reference,
                            'account': mprn,
                            'reads': reads,
                            'kwh': kwh,
                            'breakdown': breakdown,
                            'net_gbp': net,
                            'vat_gbp': vat,
                            'gross_gbp': gross,
                            'bill_type_code': bill_type_code,
                            'start_date': start_date,
                            'finish_date': finish_date,
                            'issue_date': issue_date
                        }
                    )

            elif code == "VAT":
                vatp = elements['VATP']
                if 'vat_rate' not in breakdown:
                    breakdown['vat_rate'] = set()
                breakdown['vat_rate'].add(to_decimal(vatp) / Decimal(100000))

                uvla = elements['UVLA']
                net += to_decimal(uvla) / Decimal('100')
                uvtt = elements['UVTT']
                vat += to_decimal(uvtt) / Decimal('100')
                ucsi = elements['UCSI']
                gross += to_decimal(ucsi) / Decimal('100')

            if raw_lines is not None:
                raw_lines.append(line)

        return raw_bills



import traceback
from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true
from chellow.models import GSupply, GEra, Site, SiteGEra, Session
from chellow.utils import hh_format, hh_range, req_int, req_date, csv_make_val
import chellow.computer
from chellow.computer import contract_func
import csv
from flask import g
import threading
from chellow.views import chellow_redirect
import sys
import os
from chellow.g_engine import GDataSource


def content(g_supply_id, start_date, finish_date, user):
    caches = {}
    try:
        sess = Session()
        g_supply = GSupply.get_by_id(sess, g_supply_id)

        forecast_date = chellow.computer.forecast_date()

        prev_titles = None
        running_name, finished_name = chellow.dloads.make_names(
            'g_supply_virtual_bills_hh_' + str(g_supply_id) + '.csv', user)
        f = open(running_name, mode='w', newline='')
        w = csv.writer(f, lineterminator='\n')

        for hh_start in hh_range(caches, start_date, finish_date):
            g_era = sess.query(GEra).filter(
                GEra.g_supply == g_supply, GEra.start_date <= hh_start, or_(
                    GEra.finish_date == null(),
                    GEra.finish_date >= hh_start)).one()

            site = sess.query(Site).join(SiteGEra).filter(
                SiteGEra.g_era == g_era, SiteGEra.is_physical == true()).one()

            ds = GDataSource(
                sess, hh_start, hh_start, forecast_date, g_era, caches, None)

            titles = [
                'MPRN', 'Site Code', 'Site Name', 'Account', 'HH Start', '']

            output_line = [
                ds.mprn, site.code, site.name, ds.account,
                hh_format(ds.start_date), '']

            contract = g_era.g_contract
            output_line.append('')
            contract_titles = contract_func(
                caches, contract, 'virtual_bill_titles')()
            titles.append('')
            titles.extend(contract_titles)

            contract_func(caches, contract, 'virtual_bill')(ds)
            bill = ds.bill
            for title in contract_titles:
                output_line.append(csv_make_val(bill.get(title, '')))
                if title in bill:
                    del bill[title]

            for k in sorted(bill.keys()):
                output_line.extend([k, csv_make_val(bill[k])])

            if titles != prev_titles:
                prev_titles = titles
                w.writerow(titles)
            w.writerow(output_line)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        w.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    g_supply_id = req_int('g_supply_id')
    start_date = req_date('start')
    finish_date = req_date('finish')

    args = g_supply_id, start_date, finish_date, g.user
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from collections import defaultdict
from itertools import chain

from chellow.models import (
    Batch, Bill, Channel, Contract, Era, MeasurementRequirement, Party,
    RegisterRead, Site, SiteEra, Supply, Tpr
)

from sqlalchemy import (
    false, true
)
from sqlalchemy.orm import joinedload

from werkzeug.exceptions import BadRequest


def get_era_bundles(sess, supply):
    era_bundles = []
    eras = sess.query(Era).filter(Era.supply == supply).order_by(
        Era.start_date.desc()).options(
            joinedload(Era.pc), joinedload(Era.imp_supplier_contract),
            joinedload(Era.exp_supplier_contract),
            joinedload(Era.ssc), joinedload(Era.mtc),
            joinedload(Era.mop_contract), joinedload(Era.dc_contract),
            joinedload(Era.imp_llfc), joinedload(Era.exp_llfc),
            joinedload(Era.supply).joinedload(Supply.dno)).all()
    for era in eras:
        imp_mpan_core = era.imp_mpan_core
        exp_mpan_core = era.exp_mpan_core
        physical_site = sess.query(Site).join(SiteEra).filter(
            SiteEra.is_physical == true(), SiteEra.era == era).one()
        other_sites = sess.query(Site).join(SiteEra).filter(
            SiteEra.is_physical != true(), SiteEra.era == era).all()
        imp_channels = sess.query(Channel).filter(
            Channel.era == era, Channel.imp_related == true()).order_by(
            Channel.channel_type).all()
        exp_channels = sess.query(Channel).filter(
            Channel.era == era, Channel.imp_related == false()).order_by(
            Channel.channel_type).all()
        era_bundle = {
            'era': era, 'physical_site': physical_site,
            'other_sites': other_sites, 'imp_channels': imp_channels,
            'exp_channels': exp_channels, 'imp_bills': {'bill_dicts': []},
            'exp_bills': {'bill_dicts': []},
            'dc_bills': {'bill_dicts': []}, 'mop_bills': {'bill_dicts': []}}
        era_bundles.append(era_bundle)

        if imp_mpan_core is not None:
            era_bundle['imp_shared_supplier_accounts'] = \
                sess.query(Supply).distinct().join(Era).filter(
                    Supply.id != supply.id,
                    Era.imp_supplier_account == era.imp_supplier_account,
                    Era.imp_supplier_contract == era.imp_supplier_contract) \
                .all()
        if exp_mpan_core is not None:
            era_bundle['exp_shared_supplier_accounts'] = \
                sess.query(Supply).join(Era).filter(
                    Era.supply != supply,
                    Era.exp_supplier_account == era.exp_supplier_account,
                    Era.exp_supplier_contract == era.exp_supplier_contract) \
                .all()
        if era.pc.code != '00':
            inner_headers = [
                tpr for tpr in sess.query(Tpr).join(MeasurementRequirement)
                .filter(
                    MeasurementRequirement.ssc == era.ssc).order_by(Tpr.code)]
            if era.pc.code in ['05', '06', '07', '08']:
                inner_headers.append(None)
            era_bundle['imp_bills']['inner_headers'] = inner_headers
            inner_header_codes = [
                tpr.code if tpr is not None else 'md' for tpr in inner_headers]

        bills = sess.query(Bill).filter(Bill.supply == supply).order_by(
            Bill.start_date.desc(), Bill.issue_date.desc(),
            Bill.reference.desc()).options(
            joinedload(Bill.batch).joinedload(Batch.contract).
            joinedload(Contract.party).joinedload(Party.market_role),
            joinedload(Bill.bill_type))
        if era.finish_date is not None and era != eras[0]:
            bills = bills.filter(Bill.start_date <= era.finish_date)
        if era != eras[-1]:
            bills = bills.filter(Bill.start_date >= era.start_date)

        num_outer_cols = 0
        for bill in bills:
            bill_contract = bill.batch.contract
            bill_role_code = bill_contract.party.market_role.code
            if bill_role_code == 'X':
                if exp_mpan_core is not None and \
                        bill_contract == era.exp_supplier_contract:
                    bill_group_name = 'exp_bills'
                else:
                    bill_group_name = 'imp_bills'

            elif bill_role_code in ('C', 'D'):
                bill_group_name = 'dc_bills'
            elif bill_role_code == 'M':
                bill_group_name = 'mop_bills'
            else:
                raise BadRequest(
                    f"Bill group name not found for bill_contract_id "
                    f"{bill_contract.id}.")

            bill_group = era_bundle[bill_group_name]
            rows_high = 1
            bill_dict = {'bill': bill}
            bill_group['bill_dicts'].append(bill_dict)

            if bill_group_name == 'imp_bills' and era.pc.code != '00':
                inner_tpr_map = dict((code, []) for code in inner_header_codes)
                outer_tpr_map = defaultdict(list)

                for read, tpr in sess.query(
                        RegisterRead, Tpr).join(Tpr).filter(
                        RegisterRead.bill == bill).order_by(
                        Tpr.id, RegisterRead.present_date.desc()).options(
                        joinedload(RegisterRead.previous_type),
                        joinedload(RegisterRead.present_type),
                        joinedload(RegisterRead.tpr)):
                    tpr_code = 'md' if tpr is None else tpr.code
                    try:
                        inner_tpr_map[tpr_code].append(read)
                    except KeyError:
                        outer_tpr_map[tpr_code].append(read)

                rows_high = max(
                    chain(
                        map(
                            len, chain(
                                inner_tpr_map.values(),
                                outer_tpr_map.values())),
                        [rows_high]))

                read_rows = []
                bill_dict['read_rows'] = read_rows

                for i in range(rows_high):
                    inner_reads = []
                    row_dict = {'inner_reads': inner_reads, 'outer_reads': []}
                    read_rows.append(row_dict)
                    for tpr_code in inner_header_codes:
                        try:
                            inner_reads.append(inner_tpr_map[tpr_code][i])
                        except IndexError:
                            row_dict['inner_reads'].append(None)

                    for tpr_code, read_list in outer_tpr_map.items():
                        try:
                            row_dict['outer_reads'].append(read_list[i])
                        except IndexError:
                            row_dict['outer_reads'].append(None)

                num_outer_cols = max(num_outer_cols, len(outer_tpr_map))

                bill_dict['rows_high'] = rows_high

        era_bundle['imp_bills']['num_outer_cols'] = num_outer_cols
        era_bundle['exp_bills']['num_outer_cols'] = 0

        for bill_group_name in (
                'imp_bills', 'exp_bills', 'dc_bills', 'mop_bills'):
            b_dicts = list(reversed(era_bundle[bill_group_name]['bill_dicts']))
            for i, b_dict in enumerate(b_dicts):
                if i < (len(b_dicts) - 1):
                    bill = b_dict['bill']
                    next_b_dict = b_dicts[i+1]
                    next_bill = next_b_dict['bill']
                    if (
                            bill.start_date, bill.finish_date, bill.kwh,
                            bill.net, bill.vat) == (
                            next_bill.start_date, next_bill.finish_date,
                            -1 * next_bill.kwh, -1 * next_bill.net,
                            next_bill.vat) and not (
                            (
                                bill.kwh, bill.net, bill.vat) == (0, 0, 0)
                            ) and 'collapsible' not in b_dict:
                        b_dict['collapsible'] = True
                        next_b_dict['first_collapsible'] = True
                        next_b_dict['collapsible'] = True
                        b_dict['collapse_id'] = next_b_dict['collapse_id'] = \
                            bill.id
    return era_bundles



from dateutil.relativedelta import relativedelta
from sqlalchemy import func
from sqlalchemy.sql.expression import true
from chellow.utils import (
    hh_format, HH, utc_datetime, get_file_rates, to_ct, to_utc, ct_datetime)
from werkzeug.exceptions import BadRequest
from chellow.models import HhDatum, Channel, Era
from datetime import datetime as Datetime, timedelta as Timedelta


BANDS = ('red', 'amber', 'green')

KEYS = dict(
    (
        band, {
            'kwh': 'duos-' + band + '-kwh',
            'tariff-rate': band + '-gbp-per-kwh',
            'bill-rate': 'duos-' + band + '-rate',
            'gbp': 'duos-' + band + '-gbp'}) for band in BANDS)

VL_LOOKUP = {
    'HV': {True: 'hv', False: 'hv'},
    'LV': {True: 'lv-sub', False: 'lv-net'}}


def datum_beginning_22(ds, hh):
    rs = get_file_rates(ds.caches, ds.dno_code, hh['start-date'])
    try:
        tariff = rs['tariffs'][ds.llfc_code]
        lafs = rs['lafs'][VL_LOOKUP[ds.voltage_level_code][ds.is_substation]]
    except KeyError as e:
        raise BadRequest(str(e))

    if ds.is_import:
        try:
            day_rate = float(tariff['day-gbp-per-kwh'])
        except KeyError as e:
            raise BadRequest(str(e))

        night_rate = float(tariff['night-gbp-per-kwh'])
        if 6 < hh['ct-decimal-hour'] <= 23:
            hh['duos-day-kwh'] = hh['msp-kwh']
            hh['duos-day-gbp'] = hh['msp-kwh'] * day_rate
        else:
            hh['duos-night-kwh'] = hh['msp-kwh']
            hh['duos-night-gbp'] = hh['msp-kwh'] * night_rate

    if 23 < hh['ct-decimal-hour'] <= 6:
        slot_name = 'night'
    elif hh['ct-day-of-week'] < 5 and (
            hh['ct-month'] > 10 or hh['ct-month'] < 3):
        if 15.5 < hh['ct-decimal-hour'] < 18:
            slot_name = 'winter-weekday-peak'
        elif 6 < hh['ct-decimal-hour'] < 15:
            slot_name = 'winter-weekday-day'
        else:
            slot_name = 'other'
    else:
        slot_name = 'other'

    try:
        hh['laf'] = float(lafs[slot_name])
    except KeyError as e:
        raise BadRequest(str(e))

    hh['gsp-kwh'] = hh['laf'] * hh['msp-kwh']
    hh['gsp-kw'] = hh['gsp-kwh'] * 2

    if hh['ct-is-month-end']:
        month_to = hh['start-date']
        month_from = month_to - relativedelta(months=1) + HH
        days_in_month = 0
        md_kva = 0
        month_imp_kvarh = 0
        month_kwh = 0
        for dsc in ds.get_data_sources(month_from, month_to):
            for h in dsc.hh_data:
                if h['ct-decimal-hour'] == 0:
                    days_in_month += 1
                md_kva = max(
                    md_kva,
                    (h['msp-kw'] ** 2 + h['imp-msp-kvar'] ** 2) ** 0.5)
                month_imp_kvarh += h['imp-msp-kvarh']
                month_kwh += h['msp-kwh']

        tariff = get_file_rates(
            ds.caches, ds.dno_code, hh['start-date'])['tariffs'][ds.llfc_code]
        try:
            reactive_rate = float(tariff['reactive-gbp-per-kvarh'])
        except KeyError as e:
            raise BadRequest(str(e))
        hh['duos-reactive-rate'] = reactive_rate

        if not ds.is_displaced:
            hh['duos-availability-kva'] = ds.sc
            hh['duos-excess-availability-kva'] = max(md_kva - ds.sc, 0)
            for prefix in ['', 'excess-']:
                tariff_key = prefix + 'gbp-per-kva-per-day'
                if tariff_key in tariff:
                    rate_key = 'duos-' + prefix + 'availability-rate'
                    hh[rate_key] = float(tariff[tariff_key])
                    hh['duos-' + prefix + 'availability-days'] = days_in_month
                    hh['duos-' + prefix + 'availability-gbp'] = \
                        hh[rate_key] * \
                        hh['duos-' + prefix + 'availability-kva'] * \
                        hh['duos-' + prefix + 'availability-days']

        hh['duos-reactive-gbp'] = max(0, month_imp_kvarh - month_kwh / 2) \
            * reactive_rate


def datum_beginning_20(ds, hh):
    tariff = None
    for k, tf in get_file_rates(
            ds.caches, ds.dno_code, hh['start-date'])['tariffs'].items():
        if ds.llfc_code in [cd.strip() for cd in k.split(',')]:
            tariff = tf

    if tariff is None:
        raise BadRequest(
            "The tariff for the LLFC " + ds.llfc_code +
            " cannot be found for the DNO 20 at " +
            hh_format(hh['start-date']) + ".")

    lafs = get_file_rates(
        ds.caches, ds.dno_code,
        hh['start-date'])['lafs'][ds.voltage_level_code.lower()]

    try:
        day_rate = float(tariff['day-gbp-per-kwh'])
    except KeyError as e:
        raise BadRequest(str(e))

    if 'night-gbp-per-kwh' in tariff:
        night_rate = float(tariff['night-gbp-per-kwh'])
        if 0 < hh['ct-decimal-hour'] <= 7:
            hh['duos-night-kwh'] = hh['msp-kwh']
            hh['duos-night-gbp'] = hh['msp-kwh'] * night_rate
        else:
            hh['duos-day-kwh'] = hh['msp-kwh']
            hh['duos-day-gbp'] = hh['msp-kwh'] * day_rate
    else:
        hh['duos-day-kwh'] = hh['msp-kwh']
        hh['duos-day-gbp'] = hh['msp-kwh'] * day_rate

    if 0 < hh['ct-decimal-hour'] <= 7:
        slot_name = 'night'
    elif hh['ct-day-of-week'] < 5 and \
            16 < hh['ct-decimal-hour'] <= 19 and \
            (hh['ct-month'] > 10 or hh['ct-month'] < 3):
        slot_name = 'peak'
    elif 7 > hh['ct-day-of-week'] > 1 and (
            7 < hh['ct-decimal-hour'] < 15 or
            18.5 < hh['ct-decimal-hour'] < 19) and \
            (hh['ct-month'] > 11 or hh['ct-month'] < 4):
        slot_name = 'winter-weekday'
    else:
        slot_name = 'other'
    hh['laf'] = float(lafs[slot_name])
    hh['gsp-kwh'] = hh['laf'] * hh['msp-kwh']
    hh['gsp-kw'] = hh['gsp-kwh'] * 2

    if hh['ct-is-month-end']:
        tariff = None
        for k, tf in get_file_rates(
                ds.caches, ds.dno_code, hh['start-date'])['tariffs'].items():
            if ds.llfc_code in map(str.strip, k.split(',')):
                tariff = tf
                break
        if tariff is None:
            raise BadRequest(
                "The tariff for the LLFC " + ds.llfc_code +
                " cannot be found for the DNO 20 at " +
                hh_format(hh['start-date']) + ".")
        if not ds.is_displaced:
            year_md_kva_095 = year_md_095(ds, ds.finish_date)

            hh['duos-excess-availability-kva'] = max(
                year_md_kva_095 - ds.sc, 0)
            billed_avail = max(ds.sc, year_md_kva_095)
            hh['duos-availability-kva'] = ds.sc

            for threshold, block in [
                    (15, 15), (100, 5), (250, 10), (500, 25), (1000, 50),
                    (None, 100)]:
                if threshold is None or billed_avail < threshold:
                    if billed_avail % block > 0:
                        billed_avail = (int(billed_avail / block) + 1) * block
                    break
            try:
                le_200_avail_rate = float(
                    tariff['capacity-<=200-gbp-per-kva-per-month'])
            except KeyError as e:
                raise BadRequest(str(e))

            hh['duos-availability-gbp'] = min(200, billed_avail) * \
                le_200_avail_rate

            if billed_avail > 200:
                try:
                    gt_200_avail_rate = float(
                        tariff['capacity->200-gbp-per-kva-per-month'])
                except KeyError as e:
                    raise BadRequest(str(e))

                hh['duos-availability-gbp'] += (billed_avail - 200) * \
                    gt_200_avail_rate

        try:
            if 'fixed-gbp-per-month' in tariff:
                hh['duos-standing-gbp'] = float(tariff['fixed-gbp-per-month'])
            else:
                hh['duos-standing-gbp'] = float(
                    tariff['fixed-gbp-per-day']) * hh['utc-day']
        except KeyError as e:
            raise BadRequest(str(e))


def year_md_095(data_source, finish):
    if data_source.site is None:
        return year_md_095_supply(data_source, finish)
    else:
        return year_md_095_site(data_source, finish)


def year_md_095_supply(ds, finish):
    supply = ds.supply
    sess = ds.sess
    md_kva = 0
    month_finish = finish - relativedelta(months=11)

    while not month_finish > finish:
        month_start = month_finish - relativedelta(months=1) + HH
        month_kwh_result = sess.query(
            func.sum(HhDatum.value),
            func.max(HhDatum.value)).join(Channel).join(Era).filter(
            Era.supply == supply, HhDatum.start_date >= month_start,
            HhDatum.start_date <= month_finish,
            Channel.channel_type == 'ACTIVE',
            Channel.imp_related == true()).one()

        if month_kwh_result[0] is not None:
            month_md_kw = float(month_kwh_result[1]) * 2
            month_kwh = float(month_kwh_result[0])
            month_kvarh = sess.query(
                func.sum(HhDatum.value)).join(Channel).join(Era).filter(
                Era.supply == supply, HhDatum.start_date >= month_start,
                HhDatum.start_date <= month_finish,
                Channel.channel_type == 'REACTIVE_IMP',
                Channel.imp_related == true()).one()[0]
            if month_kvarh is None:
                pf = 0.95
            else:
                month_kvarh = float(month_kvarh)
                if month_kwh == 0 and month_kvarh == 0:
                    pf = 1
                else:
                    pf = month_kwh / (month_kwh ** 2 + month_kvarh ** 2) ** 0.5
            month_kva = month_md_kw / pf
            md_kva = max(md_kva, month_kva)
        month_finish += relativedelta(months=1)
    return md_kva


def year_md_095_site(data_source, finish, pw):
    md_kva = 0
    month_finish = finish - relativedelta(months=12)

    while not month_finish > finish:
        month_start = month_finish - relativedelta(months=1) + HH
        month_data = {'start': month_start, 'finish': month_finish}
        data_source.sum_md(month_data, pw)
        if month_data['sum-kwh'] is None:
            month_md_kw = 0
            month_kwh = 0
        else:
            month_md_kw = month_data['md-kw']
            month_kwh = month_data['sum-kwh']

        data_source.sum_md(month_data, pw, False)
        if month_data['sum-kvarh'] is not None:
            month_kvarh = month_data['sum-kvarh']

        if month_kvarh == 0:
            month_kva = month_md_kw / 0.95
        else:
            if month_kwh == 0 and month_kvarh == 0:
                pf = 1
            else:
                pf = month_kwh / (month_kwh ** 2 + month_kvarh ** 2) ** 0.5
            month_kva = month_md_kw / pf
        md_kva = max(md_kva, month_kva)
        month_finish += relativedelta(months=1)
    return md_kva, ''


def datum_beginning_14(ds, hh):
    rates = get_file_rates(ds.caches, ds.dno_code, hh['start-date'])
    try:
        tariff = rates['tariffs'][ds.llfc_code]
    except KeyError as e:
        raise BadRequest(str(e))

    if 0 < hh['ct-decimal-hour'] <= 7:
        hh['duos-night-kwh'] = hh['msp-kwh']
        hh['duos-night-gbp'] = hh['msp-kwh'] * float(
            tariff['night-gbp-per-kwh'])
    else:
        hh['duos-day-kwh'] = hh['msp-kwh']
        hh['duos-day-gbp'] = hh['msp-kwh'] * float(tariff['day-gbp-per-kwh'])

    if 0 < hh['ct-decimal-hour'] <= 7:
        slot = 'night'
    elif hh['ct-day-of-week'] < 5 and hh['ct-decimal-hour'] > 15.5 and \
            hh['ct-decimal-hour'] < 19 and (
                hh['ct-month'] > 11 or hh['ct-month'] < 4):
        slot = 'winter-weekday-peak'
    elif hh['ct-day-of-week'] < 5 and (
            7 <= hh['ct-decimal-hour'] < 16 or
            18 < hh['ct-decimal-hour'] < 20) and \
            (hh['ct-month'] > 11 or hh['ct-month'] < 4):
        slot = 'winter-weekday-day'
    else:
        slot = 'other'

    hh['laf'] = float(rates['lafs'][ds.voltage_level_code.lower()][slot])
    hh['gsp-kwh'] = hh['msp-kwh'] * hh['laf']
    hh['gsp-kw'] = hh['gsp-kwh'] * 2

    if hh['utc-decimal-hour'] == 0:
        hh['duos-standing-gbp'] = float(tariff['fixed-gbp-per-day'])

    if hh['ct-is-month-end']:
        month_to = hh['start-date']
        month_from = month_to - relativedelta(months=1) + HH
        availability = ds.sc
        reactive_rate = float(tariff['reactive-gbp-per-kvarh'])
        hh['duos-reactive-rate'] = reactive_rate
        imp_msp_kvarh = 0
        msp_kwh = 0
        md_kva = 0
        for dsc in ds.get_data_sources(month_from, month_to):
            for h in dsc.hh_data:
                imp_msp_kvarh += h['imp-msp-kvarh']
                msp_kwh += h['msp-kwh']
                md_kva = max(
                    md_kva, (
                        h['msp-kw'] ** 2 + (
                            h['imp-msp-kvar'] + h['exp-msp-kvar']) ** 2) **
                    0.5)
        hh['duos-reactive-gbp'] = max(
            0, imp_msp_kvarh - msp_kwh / 3) * reactive_rate
        if not ds.is_displaced:
            availability_rate = float(
                tariff['availability-gbp-per-kva-per-day'])
            hh['duos-availability-rate'] = availability_rate
            billed_avail = max(availability, md_kva)
            hh['duos-availability-gbp'] = availability_rate * billed_avail
            hh['duos-availability-agreed-kva'] = ds.sc
            hh['duos-availability-billed-kva'] = billed_avail


def datum_2010_04_01(ds, hh):
    start_date = hh['start-date']
    dno_cache = ds.caches['dno'][ds.dno_code]

    if not ds.full_channels and not (
            hh['msp-kwh'] > 0 and hh['anti-msp-kwh'] == 0):
        imp_msp_kvarh, exp_msp_kvarh = 0, 0
    else:
        imp_msp_kvarh, exp_msp_kvarh = hh['imp-msp-kvarh'], hh['exp-msp-kvarh']

    try:
        gsp_group_cache = dno_cache[ds.gsp_group_code]
    except KeyError:
        gsp_group_cache = dno_cache[ds.gsp_group_code] = {}

    try:
        tariff = gsp_group_cache['tariffs'][ds.llfc_code][start_date]
    except KeyError:
        try:
            tariff_cache = gsp_group_cache['tariffs']
        except KeyError:
            tariff_cache = gsp_group_cache['tariffs'] = {}

        try:
            tariffs = tariff_cache[ds.llfc_code]
        except KeyError:
            tariffs = tariff_cache[ds.llfc_code] = {}

        try:
            tariff = tariffs[start_date]
        except KeyError:
            tariff = None
            try:
                tariff_list = get_file_rates(
                    ds.caches, ds.dno_code,
                    start_date)[ds.gsp_group_code]['tariffs']
            except KeyError as e:
                raise BadRequest(str(e))

            for llfcs_pcs, tf in tariff_list.items():
                key = llfcs_pcs.split('_')
                llfcs = [v.strip() for v in key[0].split(',')]
                if len(key) == 2:
                    pcs = [v.strip() for v in key[1].split(',')]
                else:
                    pcs = None

                if ds.llfc_code in llfcs and (
                        pcs is None or ds.pc_code in pcs):
                    tariff = tf
                    break

            if tariff is None:
                raise BadRequest(
                    "For the DNO " + ds.dno_code + " and timestamp " +
                    hh_format(start_date) + " and GSP group " +
                    ds.gsp_group_code + ", the LLFC '" + ds.llfc_code +
                    "' can't be found in the 'tariffs' section.")

            tariffs[start_date] = tariff

    try:
        band = gsp_group_cache['bands'][start_date]
    except KeyError:
        try:
            bands_cache = gsp_group_cache['bands']
        except KeyError:
            bands_cache = gsp_group_cache['bands'] = {}

        try:
            band = bands_cache[start_date]
        except KeyError:
            band = 'green'
            ct_hr = hh['ct-decimal-hour']
            weekend = hh['ct-day-of-week'] > 4
            try:
                slots = get_file_rates(
                    ds.caches, ds.dno_code,
                    start_date)[ds.gsp_group_code]['bands']
            except KeyError as e:
                raise BadRequest(str(e))

            for slot in slots:
                slot_weekend = slot['weekend'] == 1
                if slot_weekend == weekend and \
                        slot['start'] <= ct_hr < slot['finish']:
                    band = slot['band']
                    break

            bands_cache[start_date] = band

    try:
        laf = dno_cache['lafs'][ds.llfc_code][start_date]
    except KeyError:
        try:
            laf_cache = dno_cache['lafs']
        except KeyError:
            laf_cache = dno_cache['lafs'] = {}

        try:
            laf_cache_llfc = laf_cache[ds.llfc_code]
        except KeyError:
            laf_cache_llfc = laf_cache[ds.llfc_code] = {}

        try:
            laf = laf_cache_llfc[start_date]
        except KeyError:
            rs = get_file_rates(ds.caches, 'lafs_' + ds.dno_code, start_date)
            hist_date = rs['hist_dates'][start_date]
            try:
                hist_map_llfcs = rs._storage['hist_map']
            except KeyError:
                hist_map_llfcs = rs._storage['hist_map'] = {}

            try:
                hist_map = hist_map_llfcs[ds.llfc_code]
            except KeyError:
                hist_map = hist_map_llfcs[ds.llfc_code] = {}

            try:
                laf = hist_map[hist_date]
            except KeyError:
                try:
                    tp_id = rs['llfc_tp'][ds.llfc_code]
                except KeyError as e:
                    raise BadRequest(str(e))

                for chunk in rs['tps'][tp_id].values():
                    chunk_start_raw = Datetime.strptime(
                        chunk['start_date'], "%Y%m%d")
                    chunk_finish_raw = Datetime.strptime(
                        chunk['finish_date'], "%Y%m%d")
                    day_start_raw = chunk_start_raw
                    while day_start_raw <= chunk_finish_raw:
                        day_start_ct = to_ct(day_start_raw)
                        day_start = to_utc(day_start_ct)
                        for slot in chunk['slots']:
                            for i in range(
                                    slot['slot_start'] - 1,
                                    slot['slot_finish']):
                                dt = day_start + Timedelta(minutes=30*i)
                                hist_map[dt] = float(slot['laf'])

                        day_start_raw += Timedelta(days=1)

                laf = hist_map[hist_date]
            laf_cache_llfc[start_date] = laf

    hh['laf'] = laf
    hh['gsp-kwh'] = laf * hh['msp-kwh']
    hh['gsp-kw'] = hh['gsp-kwh'] * 2

    kvarh = max(
        max(imp_msp_kvarh, exp_msp_kvarh) -
        (0.95 ** -2 - 1) ** 0.5 * hh['msp-kwh'], 0)

    hh['duos-reactive-kvarh'] = kvarh

    duos_reactive_rate = tariff['gbp-per-kvarh']
    if duos_reactive_rate is not None:
        duos_reactive_rate = float(duos_reactive_rate)
        if duos_reactive_rate != 0:
            hh['duos-reactive-rate'] = duos_reactive_rate
            hh['duos-reactive-gbp'] = kvarh * duos_reactive_rate

    rate = float(tariff[KEYS[band]['tariff-rate']])
    hh[KEYS[band]['bill-rate']] = rate
    hh[KEYS[band]['kwh']] = hh['msp-kwh']
    hh[KEYS[band]['gbp']] = rate * hh['msp-kwh']

    if hh['ct-decimal-hour'] == 23.5 and not ds.is_displaced:
        hh['duos-fixed-days'] = 1
        rate = float(tariff['gbp-per-mpan-per-day'])
        hh['duos-fixed-rate'] = rate
        hh['duos-fixed-gbp'] = rate

        hh['duos-availability-days'] = 1
        kva = ds.sc
        hh['duos-availability-kva'] = kva
        rate = float(tariff['gbp-per-kva-per-day'])
        hh['duos-availability-rate'] = rate
        hh['duos-availability-gbp'] = rate * kva

    if hh['ct-is-month-end'] and not ds.is_displaced:
        month_to = start_date
        month_from = to_utc(ct_datetime(hh['ct-year'], hh['ct-month'], 1))
        md_kva = 0
        days_in_month = 0
        for dsc in ds.get_data_sources(month_from, month_to):
            for datum in dsc.hh_data:
                md_kva = max(
                    md_kva, (
                        datum['msp-kw'] ** 2 + max(
                            datum['imp-msp-kvar'], datum['exp-msp-kvar']) **
                        2) ** 0.5)
                if datum['ct-decimal-hour'] == 0:
                    days_in_month += 1

        excess_kva = max(md_kva - ds.sc, 0)

        if 'excess-gbp-per-kva-per-day' in tariff and excess_kva != 0:
            rate = float(tariff['excess-gbp-per-kva-per-day'])
            hh['duos-excess-availability-kva'] = excess_kva
            rate = float(tariff['excess-gbp-per-kva-per-day'])
            hh['duos-excess-availability-rate'] = rate
            hh['duos-excess-availability-days'] = days_in_month
            hh['duos-excess-availability-gbp'] = rate * excess_kva * \
                days_in_month


def datum_2012_02_23(ds, hh):
    start_date = hh['start-date']
    dno_cache = ds.caches['dno'][ds.dno_code]

    if not ds.full_channels and hh['msp-kwh'] == 0:
        imp_msp_kvarh, exp_msp_kvarh = 0, 0
    else:
        imp_msp_kvarh, exp_msp_kvarh = hh['imp-msp-kvarh'], hh['exp-msp-kvarh']

    try:
        gsp_group_cache = dno_cache[ds.gsp_group_code]
    except KeyError:
        gsp_group_cache = dno_cache[ds.gsp_group_code] = {}

    try:
        tariff = gsp_group_cache['tariffs'][ds.pc_code][ds.llfc_code][
            start_date]
    except KeyError:
        try:
            tariffs_cache = gsp_group_cache['tariffs']
        except KeyError:
            tariffs_cache = gsp_group_cache['tariffs'] = {}

        try:
            pc_cache = tariffs_cache[ds.pc_code]
        except KeyError:
            pc_cache = tariffs_cache[ds.pc_code] = {}

        try:
            tariffs = pc_cache[ds.llfc_code]
        except KeyError:
            tariffs = pc_cache[ds.llfc_code] = {}

        try:
            tariff = tariffs[start_date]
        except KeyError:
            tariff = None
            try:
                tariff_list = get_file_rates(
                    ds.caches, ds.dno_code,
                    start_date)[ds.gsp_group_code]['tariffs']
            except KeyError as e:
                raise BadRequest(str(e))

            for llfcs_pcs, tf in tariff_list.items():
                key = llfcs_pcs.split('_')
                llfcs = [v.strip() for v in key[0].split(',')]
                if len(key) == 2:
                    pcs = [v.strip() for v in key[1].split(',')]
                else:
                    pcs = None

                if ds.llfc_code in llfcs and (
                        pcs is None or ds.pc_code in pcs):
                    tariff = tf
                    break

            if tariff is None:
                raise BadRequest(
                    "For the DNO " + ds.dno_code + " and timestamp " +
                    hh_format(start_date) + " and GSP group " +
                    ds.gsp_group_code + ", the LLFC '" + ds.llfc_code +
                    " with PC " + ds.pc_code +
                    "' can't be found in the 'tariffs' section.")

            tariffs[start_date] = tariff

    try:
        band = gsp_group_cache['bands'][start_date]
    except KeyError:
        try:
            bands_cache = gsp_group_cache['bands']
        except KeyError:
            bands_cache = gsp_group_cache['bands'] = {}

        try:
            band = bands_cache[start_date]
        except KeyError:
            band = 'green'
            ct_hr = hh['ct-decimal-hour']
            weekend = hh['ct-day-of-week'] > 4
            try:
                slots = get_file_rates(
                    ds.caches, ds.dno_code,
                    start_date)[ds.gsp_group_code]['bands']
            except KeyError as e:
                raise BadRequest(str(e))

            for slot in slots:
                slot_weekend = slot['weekend'] == 1
                if slot_weekend == weekend and \
                        slot['start'] <= ct_hr < slot['finish']:
                    band = slot['band']
                    break

            bands_cache[start_date] = band

    try:
        laf = dno_cache['lafs'][ds.llfc_code][start_date]
    except KeyError:
        try:
            laf_cache = dno_cache['lafs']
        except KeyError:
            laf_cache = dno_cache['lafs'] = {}

        try:
            laf_cache_llfc = laf_cache[ds.llfc_code]
        except KeyError:
            laf_cache_llfc = laf_cache[ds.llfc_code] = {}

        try:
            laf = laf_cache_llfc[start_date]
        except KeyError:
            rs = get_file_rates(ds.caches, 'lafs_' + ds.dno_code, start_date)
            hist_date = rs['hist_dates'][start_date]
            try:
                hist_map_llfcs = rs._storage['hist_map']
            except KeyError:
                hist_map_llfcs = rs._storage['hist_map'] = {}

            try:
                hist_map = hist_map_llfcs[ds.llfc_code]
            except KeyError:
                hist_map = hist_map_llfcs[ds.llfc_code] = {}

            try:
                laf = hist_map[hist_date]
            except KeyError:
                try:
                    tp_id = rs['llfc_tp'][ds.llfc_code]
                except KeyError as e:
                    raise BadRequest(str(e))

                for chunk in rs['tps'][tp_id].values():
                    chunk_start_raw = Datetime.strptime(
                        chunk['start_date'], "%Y%m%d")
                    chunk_finish_raw = Datetime.strptime(
                        chunk['finish_date'], "%Y%m%d")
                    day_start_raw = chunk_start_raw
                    while day_start_raw <= chunk_finish_raw:
                        day_start_ct = to_ct(day_start_raw)
                        day_start = to_utc(day_start_ct)
                        for slot in chunk['slots']:
                            for i in range(
                                    slot['slot_start'] - 1,
                                    slot['slot_finish']):
                                dt = day_start + Timedelta(minutes=30*i)
                                hist_map[dt] = float(slot['laf'])

                        day_start_raw += Timedelta(days=1)

                laf = hist_map[hist_date]
            laf_cache_llfc[start_date] = laf

    hh['laf'] = laf
    hh['gsp-kwh'] = laf * hh['msp-kwh']
    hh['gsp-kw'] = hh['gsp-kwh'] * 2

    kvarh = max(
        max(imp_msp_kvarh, exp_msp_kvarh) -
        (0.95 ** -2 - 1) ** 0.5 * hh['msp-kwh'], 0)

    hh['duos-reactive-kvarh'] = kvarh

    duos_reactive_rate = tariff['gbp-per-kvarh']
    if duos_reactive_rate is not None:
        duos_reactive_rate = float(duos_reactive_rate)
        if duos_reactive_rate != 0:
            hh['duos-reactive-rate'] = duos_reactive_rate
            hh['duos-reactive-gbp'] = kvarh * duos_reactive_rate

    rate = float(tariff[KEYS[band]['tariff-rate']])
    hh[KEYS[band]['bill-rate']] = rate
    hh[KEYS[band]['kwh']] = hh['msp-kwh']
    hh[KEYS[band]['gbp']] = rate * hh['msp-kwh']

    if hh['ct-decimal-hour'] == 23.5 and not ds.is_displaced:
        hh['duos-fixed-days'] = 1
        rate = float(tariff['gbp-per-mpan-per-day'])
        hh['duos-fixed-rate'] = rate
        hh['duos-fixed-gbp'] = rate

        hh['duos-availability-days'] = 1
        kva = ds.sc
        hh['duos-availability-kva'] = kva
        rate = float(tariff['gbp-per-kva-per-day'])
        hh['duos-availability-rate'] = rate
        hh['duos-availability-gbp'] = rate * kva

    if hh['ct-is-month-end'] and not ds.is_displaced:
        month_to = start_date
        month_from = to_utc(ct_datetime(hh['ct-year'], hh['ct-month'], 1))
        md_kva = 0
        days_in_month = 0
        for dsc in ds.get_data_sources(month_from, month_to):
            for datum in dsc.hh_data:
                md_kva = max(
                    md_kva, (
                        datum['msp-kw'] ** 2 + max(
                            datum['imp-msp-kvar'], datum['exp-msp-kvar']) **
                        2) ** 0.5)
                if datum['ct-decimal-hour'] == 0:
                    days_in_month += 1

        excess_kva = max(md_kva - ds.sc, 0)

        if 'excess-gbp-per-kva-per-day' in tariff and excess_kva != 0:
            rate = float(tariff['excess-gbp-per-kva-per-day'])
            hh['duos-excess-availability-kva'] = excess_kva
            rate = float(tariff['excess-gbp-per-kva-per-day'])
            hh['duos-excess-availability-rate'] = rate
            hh['duos-excess-availability-days'] = days_in_month
            hh['duos-excess-availability-gbp'] = rate * excess_kva * \
                days_in_month


CUTOFF_DATE_1 = utc_datetime(2010, 3, 31, 23, 0)
CUTOFF_DATE_2 = utc_datetime(2012, 2, 23, 0, 0)


def duos_vb(ds):
    try:
        data_func_cache = ds.caches['dno'][ds.dno_code]['data_funcs']
    except KeyError:
        try:
            dno_caches = ds.caches['dno']
        except KeyError:
            dno_caches = ds.caches['dno'] = {}

        try:
            dno_cache = dno_caches[ds.dno_code]
        except KeyError:
            dno_cache = dno_caches[ds.dno_code] = {}

        try:
            data_func_cache = dno_cache['data_funcs']
        except KeyError:
            data_func_cache = dno_cache['data_funcs'] = {}

    for hh in ds.hh_data:
        try:
            data_func_cache[hh['start-date']](ds, hh)
        except KeyError:
            if hh['start-date'] < CUTOFF_DATE_1:
                if ds.dno_code == '14':
                    data_func = datum_beginning_14
                elif ds.dno_code == '20':
                    data_func = datum_beginning_20
                elif ds.dno_code in ('22', '99'):
                    data_func = datum_beginning_22
                else:
                    raise Exception('Not recognized')
            elif hh['start-date'] < CUTOFF_DATE_2:
                data_func = datum_2010_04_01
            else:
                data_func = datum_2012_02_23

            data_func_cache[hh['start-date']] = data_func
            data_func(ds, hh)



from sqlalchemy import or_, and_
from sqlalchemy.orm import joinedload
import traceback
from chellow.models import (
    RegisterRead, Bill, Supply, Era, Session, Batch, BillType)
from chellow.utils import csv_make_val, c_months_u, req_int
from flask import request, g
import chellow.dloads
import csv
import os
from werkzeug.exceptions import BadRequest
import threading
from chellow.views import chellow_redirect


def content(year, month, months, supply_id, user):
    sess = f = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'register_reads.csv', user)
        f = open(running_name, mode='w', newline='')
        w = csv.writer(f, lineterminator='\n')
        titles = (
            'Duration Start', 'Duration Finish', 'Supply Id',
            'Import MPAN Core', 'Export MPAN Core', 'Batch Reference',
            'Bill Id', 'Bill Reference', 'Bill Issue Date', 'Bill Type',
            'Register Read Id', 'TPR', 'Coefficient', 'Previous Read Date',
            'Previous Read Value', 'Previous Read Type', 'Present Read Date',
            'Present Read Value', 'Present Read Type'
        )
        w.writerow(titles)

        month_pairs = list(
            c_months_u(finish_year=year, finish_month=month, months=months))
        start_date, finish_date = month_pairs[0][0], month_pairs[-1][-1]

        supplies = sess.query(Supply).join(Bill).join(RegisterRead).filter(
            or_(
                and_(
                    RegisterRead.present_date >= start_date,
                    RegisterRead.present_date <= finish_date),
                and_(
                    RegisterRead.previous_date >= start_date,
                    RegisterRead.previous_date <= finish_date))).order_by(
            Bill.supply_id).distinct()

        if supply_id is not None:
            supply = Supply.get_by_id(sess, supply_id)
            supplies = supplies.filter(Bill.supply == supply)

        for supply in supplies:
            supply_id = supply.id
            for bill, batch, bill_type in sess.query(
                    Bill, Batch, BillType).join(Batch).join(BillType).join(
                    RegisterRead).filter(
                    Bill.supply == supply, or_(
                        and_(
                            RegisterRead.present_date >= start_date,
                            RegisterRead.present_date <= finish_date),
                        and_(
                            RegisterRead.previous_date >= start_date,
                            RegisterRead.previous_date <= finish_date))):

                era = supply.find_era_at(sess, bill.start_date)
                if era is None:
                    eras = sess.query(Era).filter(
                        Era.supply == supply).order_by(Era.start_date).all()
                    if bill.start_date < eras[0].start_date:
                        era = eras[0]
                    else:
                        era = eras[-1]

                for read in sess.query(RegisterRead).filter(
                        RegisterRead.bill == bill, or_(
                            and_(
                                RegisterRead.present_date >= start_date,
                                RegisterRead.present_date <= finish_date),
                            and_(
                                RegisterRead.previous_date >= start_date,
                                RegisterRead.previous_date <= finish_date))
                        ).options(
                            joinedload(RegisterRead.tpr),
                            joinedload(RegisterRead.previous_type),
                            joinedload(RegisterRead.present_type)):

                    vals = [
                        start_date, finish_date, supply_id, era.imp_mpan_core,
                        era.exp_mpan_core, batch.reference, bill.id,
                        bill.reference, bill.issue_date, bill_type.code,
                        read.id, 'md' if read.tpr is None else read.tpr.code,
                        read.coefficient, read.previous_date,
                        read.previous_value, read.previous_type.code,
                        read.present_date, read.present_value,
                        read.present_type.code
                    ]
                    w.writerow(csv_make_val(v) for v in vals)

                # Avoid a long-running transaction
                sess.rollback()

    except BadRequest as e:
        w.writerow([e.description])
    except BaseException:
        msg = traceback.format_exc()
        f.write(msg)
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    year = req_int('end_year')
    month = req_int('end_month')
    months = req_int('months')
    supply_id = req_int('supply_id') if 'supply_id' in request.values else None
    args = (year, month, months, supply_id, g.user)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import os
import traceback
from sqlalchemy import or_, true
from sqlalchemy.sql.expression import null
from sqlalchemy.orm import joinedload
from chellow.models import (
    Session, GContract, Site, GEra, SiteGEra, GSupply, GBill)
from chellow.computer import contract_func
from chellow.g_engine import GDataSource
import chellow.computer
import chellow.dloads
import threading
import odio
import sys
from werkzeug.exceptions import BadRequest
from chellow.utils import (
    hh_format, hh_max, hh_min, req_int, req_bool, make_val, ct_datetime_now,
    c_months_u)
from flask import request, g
from chellow.views import chellow_redirect


CATEGORY_ORDER = {None: 0, 'unmetered': 1, 'nhh': 2, 'amr': 3, 'hh': 4}
meter_order = {'hh': 0, 'amr': 1, 'nhh': 2, 'unmetered': 3}


def write_spreadsheet(fl, compressed, site_rows, era_rows):
    fl.seek(0)
    fl.truncate()
    with odio.create_spreadsheet(fl, '1.2', compressed=compressed) as f:
        f.append_table("Site Level", site_rows)
        f.append_table("Era Level", era_rows)


def content(
        site_id, g_supply_id, user, compression, finish_year, finish_month,
        months):
    now = ct_datetime_now()
    report_context = {}
    sess = None
    month_list = list(
        c_months_u(
            finish_year=finish_year, finish_month=finish_month, months=months))
    start_date, finish_date = month_list[0][0], month_list[-1][-1]

    try:
        sess = Session()
        base_name = [
            'g_monthly_duration',
            hh_format(start_date).replace(' ', '_').replace(':', '').
            replace('-', ''), 'for', str(months), 'months'
        ]

        forecast_from = chellow.computer.forecast_date()

        sites = sess.query(Site).join(SiteGEra).join(GEra).filter(
            SiteGEra.is_physical == true()).distinct().order_by(Site.code)
        if site_id is not None:
            site = Site.get_by_id(sess, site_id)
            sites = sites.filter(Site.id == site.id)
            base_name.append('site')
            base_name.append(site.code)
        if g_supply_id is not None:
            g_supply = GSupply.get_by_id(sess, g_supply_id)
            base_name.append('g_supply')
            base_name.append(str(g_supply.id))
            sites = sites.filter(GEra.g_supply == g_supply)

        running_name, finished_name = chellow.dloads.make_names(
            '_'.join(base_name) + '.ods', user)

        rf = open(running_name, "wb")
        site_rows = []
        g_era_rows = []

        era_header_titles = [
            'creation_date', 'mprn', 'supply_name', 'exit_zone', 'msn', 'unit',
            'contract', 'site_id', 'site_name', 'associated_site_ids', 'month']
        site_header_titles = [
            'creation_date', 'site_id', 'site_name', 'associated_site_ids',
            'month']
        summary_titles = ['kwh', 'gbp', 'billed_kwh', 'billed_gbp']

        vb_titles = []
        conts = sess.query(GContract).join(GEra).join(GSupply).filter(
            GEra.start_date <= finish_date, or_(
                GEra.finish_date == null(),
                GEra.finish_date >= start_date)).distinct().order_by(
            GContract.id)
        if g_supply_id is not None:
            conts = conts.filter(GEra.g_supply_id == g_supply_id)
        for cont in conts:
            title_func = chellow.computer.contract_func(
                report_context, cont, 'virtual_bill_titles')
            if title_func is None:
                raise Exception(
                    "For the contract " + cont.name + " there doesn't seem " +
                    "to be a 'virtual_bill_titles' function.")
            for title in title_func():
                if title not in vb_titles:
                    vb_titles.append(title)

        g_era_rows.append(era_header_titles + summary_titles + vb_titles)
        site_rows.append(site_header_titles + summary_titles)

        for month_start, month_finish in month_list:
            for site in sites.filter(
                    GEra.start_date <= month_finish, or_(
                        GEra.finish_date == null(),
                        GEra.finish_date >= month_start)):
                site_kwh = site_gbp = site_billed_kwh = site_billed_gbp = 0

                for g_era in sess.query(GEra).join(SiteGEra).filter(
                        SiteGEra.site == site, SiteGEra.is_physical == true(),
                        GEra.start_date <= month_finish, or_(
                            GEra.finish_date == null(),
                            GEra.finish_date >= month_start)).options(
                        joinedload(GEra.g_contract), joinedload(GEra.g_supply),
                        joinedload(GEra.g_supply).joinedload(
                            GSupply.g_exit_zone)).order_by(GEra.id):

                    g_supply = g_era.g_supply

                    if g_supply_id is not None and g_supply.id != g_supply_id:
                        continue

                    ss_start = hh_max(g_era.start_date, month_start)
                    ss_finish = hh_min(g_era.finish_date, month_finish)

                    ss = GDataSource(
                        sess, ss_start, ss_finish, forecast_from, g_era,
                        report_context, None)

                    contract = g_era.g_contract
                    vb_function = contract_func(
                        report_context, contract, 'virtual_bill')
                    if vb_function is None:
                        raise BadRequest(
                            "The contract " + contract.name +
                            " doesn't have the virtual_bill() function.")
                    vb_function(ss)
                    bill = ss.bill

                    try:
                        gbp = bill['net_gbp']
                    except KeyError:
                        gbp = 0
                        bill['problem'] += 'For the supply ' + ss.mprn + \
                            ' the virtual bill ' + str(bill) + \
                            ' from the contract ' + contract.name + \
                            ' does not contain the net_gbp key.'
                    try:
                        kwh = bill['kwh']
                    except KeyError:
                        kwh = 0
                        bill['problem'] += "For the supply " + ss.mprn + \
                            " the virtual bill " + str(bill) + \
                            " from the contract " + contract.name + \
                            " does not contain the 'kwh' key."

                    billed_kwh = billed_gbp = 0

                    g_era_associates = {
                        s.site.code for s in g_era.site_g_eras
                        if not s.is_physical}

                    for g_bill in sess.query(GBill).filter(
                            GBill.g_supply == g_supply,
                            GBill.start_date <= ss_finish,
                            GBill.finish_date >= ss_start):
                        bill_start = g_bill.start_date
                        bill_finish = g_bill.finish_date
                        bill_duration = (
                            bill_finish - bill_start).total_seconds() + \
                            (30 * 60)
                        overlap_duration = (
                            min(bill_finish, ss_finish) -
                            max(bill_start, ss_start)
                            ).total_seconds() + (30 * 60)
                        overlap_proportion = overlap_duration / bill_duration
                        billed_kwh += overlap_proportion * float(g_bill.kwh)
                        billed_gbp += overlap_proportion * float(g_bill.net)

                    associated_site_ids = ','.join(sorted(g_era_associates))
                    g_era_rows.append(
                        [
                            make_val(v) for v in [
                                now, g_supply.mprn, g_supply.name,
                                g_supply.g_exit_zone.code, g_era.msn,
                                g_era.g_unit.code, contract.name, site.code,
                                site.name, associated_site_ids, month_finish,
                                kwh, gbp, billed_kwh, billed_gbp]
                        ] +
                        [make_val(bill.get(t)) for t in vb_titles]
                    )

                    site_kwh += kwh
                    site_gbp += gbp
                    site_billed_kwh += billed_kwh
                    site_billed_gbp += billed_gbp

                linked_sites = ', '.join(
                    s.code for s in site.find_linked_sites(
                        sess, month_start, month_finish))

                site_rows.append(
                    [
                        make_val(v) for v in [
                            now, site.code, site.name, linked_sites,
                            month_finish, site_kwh, site_gbp, site_billed_kwh,
                            site_billed_gbp]
                    ]
                )
                sess.rollback()
            write_spreadsheet(rf, compression, site_rows, g_era_rows)

    except BadRequest as e:
        site_rows.append(["Problem " + e.description])
        write_spreadsheet(rf, compression, site_rows, g_era_rows)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + '\n')
        site_rows.append(["Problem " + msg])
        write_spreadsheet(rf, compression, site_rows, g_era_rows)
    finally:
        if sess is not None:
            sess.close()
        try:
            rf.close()
            os.rename(running_name, finished_name)
        except BaseException:
            msg = traceback.format_exc()
            r_name, f_name = chellow.dloads.make_names('error.txt', user)
            ef = open(r_name, "w")
            ef.write(msg + '\n')
            ef.close()


def do_get(sess):
    finish_year = req_int("finish_year")
    finish_month = req_int("finish_month")
    months = req_int("months")

    site_id = req_int('site_id') if 'site_id' in request.values else None

    if 'g_supply_id' in request.values:
        g_supply_id = req_int('g_supply_id')
    else:
        g_supply_id = None

    if 'compression' in request.values:
        compression = req_bool('compression')
    else:
        compression = True

    user = g.user
    args = (
        site_id, g_supply_id, user, compression, finish_year, finish_month,
        months)

    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import decimal
import itertools
from chellow.utils import parse_mpan_core, HH, utc_datetime
from werkzeug.exceptions import BadRequest


def create_parser(reader, mpan_map):
    return StarkDf2HhParser(reader, mpan_map)


class StarkDf2HhParser():
    def __init__(self, reader, mpan_map):
        self.line = self.line_number = None
        self.reader = zip(itertools.count(1), reader)

        self.line_number, self.line = next(self.reader)
        if self.line.strip().upper() != "#F2":
            raise BadRequest("The first line must be '#F2'.")
        self.sensor_map = {
            1: 'ACTIVE', 2: 'ACTIVE', 3: 'REACTIVE_IMP', 4: 'REACTIVE_EXP'}

    def __iter__(self):
        return self

    def __next__(self):
        local_datum = None
        try:
            while local_datum is None:
                self.line_number, self.line = next(self.reader)
                lline = self.line.strip().upper()
                if lline.startswith("#O"):
                    self.core = parse_mpan_core(lline[2:])
                elif lline.startswith("#S"):
                    sensor = int(lline[2:].strip())
                    try:
                        self.channel_type = self.sensor_map[sensor]
                    except KeyError:
                        raise BadRequest(
                            "The sensor number must be between 1 and 4 "
                            "inclusive.")
                elif lline.startswith("#F2") or len(lline) == 0:
                    continue
                else:
                    fields = [f.strip() for f in lline.split(',')]
                    if len(fields) != 4:
                        raise BadRequest(
                            "There should be 4 comma separated values, but I "
                            "found " + str(len(fields)) + ".")

                    d_day, d_month, d_year = map(int, fields[0].split('/'))
                    time_fields = tuple(map(int, fields[1].split(':')))
                    if len(time_fields) > 2 and time_fields[2] != 0:
                        raise BadRequest(
                            "The number of seconds (if present) must always "
                            "be zero.")
                    start_date = utc_datetime(
                        d_year, d_month, d_day, time_fields[0],
                        time_fields[1]) - HH

                    try:
                        value = decimal.Decimal(fields[2])
                    except ValueError:
                        raise BadRequest(
                            "Problem parsing the value: " + fields[2])
                    status = fields[3][-1]
                    local_datum = {
                        'mpan_core': self.core,
                        'channel_type': self.channel_type,
                        'start_date': start_date, 'value': value,
                        'status': status}
            return local_datum
        except BadRequest as e:
            e.description = ''.join(
                (
                    "Problem at line number: ", str(self.line_number), ": ",
                    self.line, ": ", e.description))
            raise e

    def close(self):
        self.reader.close()

    def get_status(self):
        return "Reached line number: " + str(self.line_number)



from datetime import datetime as Datetime
from dateutil.relativedelta import relativedelta
import pytz
import traceback
from sqlalchemy.sql.expression import true, or_, null
from chellow.utils import (
    HH, hh_min, hh_max, hh_format, req_int, req_bool, csv_make_val)
from chellow.views import chellow_redirect
from chellow.models import Supply, Site, SiteEra, Session, Era
import chellow.computer
from flask import g
import csv
from werkzeug.exceptions import BadRequest
import os
import threading


def content(
        start_year, start_month, start_day, finish_year, finish_month,
        finish_day, is_import, supply_id, user):
    caches = {}
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'daily_supplier_virtual_bill.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')
        start_date = Datetime(
            start_year, start_month, start_day, tzinfo=pytz.utc)
        finish_date = Datetime(
            finish_year, finish_month, finish_day, tzinfo=pytz.utc) + \
            relativedelta(days=1) - HH

        supply = Supply.get_by_id(sess, supply_id)
        forecast_date = chellow.computer.forecast_date()
        day_start = start_date
        header_titles = [
            'MPAN Core', 'Site Code', 'Site Name', 'Account', 'From', 'To',
            'Is Forecast?'
        ]

        bill_titles = []
        # Find titles
        for era in sess.query(Era).filter(
                Era.supply == supply, Era.start_date <= finish_date, or_(
                    Era.finish_date == null(),
                    Era.finish_date >= start_date)):

            if is_import:
                cont = era.imp_supplier_contract
            else:
                cont = era.exp_supplier_contract

            for title in chellow.computer.contract_func(
                    caches, cont, 'virtual_bill_titles')():
                if title not in bill_titles:
                    bill_titles.append(title)

            ssc = era.ssc
            if ssc is not None:
                for mr in ssc.measurement_requirements:
                    for suffix in ('-kwh', '-rate', '-gbp'):
                        title = mr.tpr.code + suffix
                        if title not in bill_titles:
                            bill_titles.append(title)

        writer.writerow(header_titles + bill_titles)

        while not day_start > finish_date:
            day_finish = day_start + relativedelta(days=1) - HH

            for era in supply.find_eras(sess, day_start, day_finish):
                chunk_start = hh_max(era.start_date, day_start)
                chunk_finish = hh_min(era.finish_date, day_finish)

                ss = chellow.computer.SupplySource(
                    sess, chunk_start, chunk_finish, forecast_date, era,
                    is_import, caches)

                site = sess.query(Site).join(SiteEra).filter(
                    SiteEra.era == era, SiteEra.is_physical == true()).one()
                row = [
                    ss.mpan_core, site.code, site.name, ss.supplier_account,
                    hh_format(ss.start_date), hh_format(ss.finish_date),
                    ss.years_back > 0]

                chellow.computer.contract_func(
                    caches, ss.supplier_contract, 'virtual_bill')(ss)
                bill = ss.supplier_bill
                for title in bill_titles:
                    if title in bill:
                        row.append(csv_make_val(bill[title]))
                        del bill[title]
                    else:
                        row.append('')

                for k in sorted(bill.keys()):
                    row.append(k)
                    row.append(csv_make_val(bill[k]))
                writer.writerow(row)

            day_start += relativedelta(days=1)
    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    start_year = req_int('start_year')
    start_month = req_int('start_month')
    start_day = req_int("start_day")

    finish_year = req_int('finish_year')
    finish_month = req_int('finish_month')
    finish_day = req_int('finish_day')

    is_import = req_bool('is_import')
    supply_id = req_int('supply_id')

    threading.Thread(
        target=content, args=(
            start_year, start_month, start_day, finish_year, finish_month,
            finish_day, is_import, supply_id, g.user)).start()
    return chellow_redirect("/downloads", 303)



from chellow.utils import get_file_rates


DAILY_THRESHOLD = 145


def vb(ds):
    for hh in ds.hh_data:
        rate = float(
            get_file_rates(
                ds.caches, 'g_ccl', hh['start_date'])['ccl_gbp_per_kwh'])

        hh['ccl'] = rate



from sqlalchemy import or_
from sqlalchemy.sql.expression import null
import traceback
from chellow.utils import (
    HH, csv_make_val, req_int, ct_datetime, reduce_bill_hhs, to_utc,
    c_months_u)
from chellow.models import Site, SiteEra, Era, Supply, Source, Session, Pc
import chellow.computer
import chellow.duos
import chellow.triad
from flask import request, g
import csv
import chellow.dloads
import os
from chellow.views import chellow_redirect
import threading
from werkzeug.exceptions import BadRequest


def _make_sites(sess, year_start, year_finish, site_id, source_codes):
    sites = sess.query(Site).join(SiteEra).join(Era).join(Pc).join(
        Supply).join(Source).filter(
        Pc.code == '00',
        Source.code.in_(source_codes),
        Era.start_date <= year_finish, or_(
            Era.finish_date == null(),
            Era.finish_date >= year_start)).distinct().order_by(Site.code)

    if site_id is not None:
        sites = sites.filter(Site.id == site_id)

    return sites


def _write_sites(sess, caches, writer, year, site_id):
    titles = (
        "Site Code", "Site Name", "Displaced TRIAD 1 Date",
        "Displaced TRIAD 1 MSP kW", "Displaced TRIAD LAF",
        "Displaced TRIAD 1 GSP kW", "Displaced TRIAD 2 Date",
        "Displaced TRIAD 2 MSP kW", "Displaced TRIAD 2 LAF",
        "Displaced TRIAD 2 GSP kW", "Displaced TRIAD 3 Date",
        "Displaced TRIAD 3 MSP kW", "Displaced TRIAD 3 LAF",
        "Displaced TRIAD 3 GSP kW", "Displaced GSP kW",
        "Displaced Rate GBP / kW", "GBP"
    )
    writer.writerow(titles)

    march_finish_ct = ct_datetime(year, 4, 1) - HH
    march_finish_utc = to_utc(march_finish_ct)
    march_start_ct = ct_datetime(year, 3, 1)
    march_start_utc = to_utc(march_start_ct)
    year_start = to_utc(ct_datetime(year - 1, 4, 1))

    forecast_date = chellow.computer.forecast_date()

    sites = _make_sites(
        sess, year_start, march_finish_utc, site_id, ('gen', 'gen-net'))

    scalar_names = {
        'triad-actual-gsp-kw', 'triad-actual-gbp'
    }

    rate_names = {
        'triad-actual-rate', 'triad-estimate-rate'
    }

    for i in range(1, 4):
        pref = 'triad-actual-' + str(i) + '-'
        for suf in ('msp-kw', 'gsp-kw'):
            scalar_names.add(pref + suf)
        for suf in ('date', 'status', 'laf'):
            rate_names.add(pref + suf)

    for site in sites:
        displaced_era = None
        for month_start, month_finish in sorted(
                c_months_u(start_year=year - 1, start_month=4, months=12),
                reverse=True):
            displaced_era = chellow.computer.displaced_era(
                sess, caches, site, month_start, month_finish, forecast_date)
            if displaced_era is not None:
                break

        if displaced_era is None:
            break

        site_ds = chellow.computer.SiteSource(
            sess, site, march_start_utc, march_finish_utc, forecast_date,
            caches, displaced_era)
        chellow.duos.duos_vb(site_ds)
        chellow.triad.hh(site_ds)

        for hh in site_ds.hh_data:
            bill_hh = site_ds.supplier_bill_hhs[hh['start-date']]
            for k in scalar_names & hh.keys():
                bill_hh[k] = hh[k]

            for k in rate_names & hh.keys():
                bill_hh[k] = {hh[k]}

        bill = reduce_bill_hhs(site_ds.supplier_bill_hhs)
        values = [site.code, site.name]
        for i in range(1, 4):
            triad_prefix = 'triad-actual-' + str(i) + '-'
            for suffix in ('date', 'msp-kw', 'laf', 'gsp-kw'):
                values.append(csv_make_val(bill[triad_prefix + suffix]))

        for suffix in ('gsp-kw', 'rate', 'gbp'):
            values.append(csv_make_val(bill['triad-actual-' + suffix]))

        writer.writerow(values)

        # Avoid long-running transaction
        sess.rollback()


def content(year, site_id, user):
    caches = {}
    sess = f = writer = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'output.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')
        _write_sites(sess, caches, writer, year, site_id)

    except BadRequest as e:
        writer.writerow([e.description])
    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    site_id = req_int('site_id') if 'site_id' in request.values else None
    year = req_int('year')
    threading.Thread(target=content, args=(year, site_id, g.user)).start()
    return chellow_redirect("/downloads", 303)



from chellow.utils import get_file_rates


def hh(supply_source):
    for hh in supply_source.hh_data:
        hh['aahedc-kwh'] = hh['gsp-kwh']
        rate = float(
            get_file_rates(
                supply_source.caches, 'aahedc',
                hh['start-date'])['aahedc_gbp_per_gsp_kwh'])
        hh['aahedc-rate'] = rate
        hh['aahedc-gbp'] = hh['gsp-kwh'] * rate



from decimal import Decimal, InvalidOperation
from chellow.utils import parse_mpan_core, to_utc, ct_datetime
from xlrd import xldate_as_tuple, open_workbook
from werkzeug.exceptions import BadRequest
from chellow.models import Session
from datetime import datetime as Datetime


def get_date_ct(row, name, datemode):
    val = get_value(row, name)
    if isinstance(val, float):
        return to_utc(Datetime(*xldate_as_tuple(val, datemode)))


def get_start_date(row, name, datemode):
    return to_utc(get_date_ct(row, name, datemode))


def get_finish_date(row, name, datemode):
    d = get_date_ct(row, name, datemode)
    return to_utc(ct_datetime(d.year, d.month, d.day, 23, 30))


def get_value(row, idx):
    try:
        return row[idx].value
    except IndexError:
        raise BadRequest(
            "For the row " + str(row) + ", the index is " + str(idx) +
            " which is beyond the end of the row. ")


def get_str(row, idx):
    return get_value(row, idx).strip()


def get_dec(row, idx):
    try:
        return Decimal(str(get_value(row, idx)))
    except InvalidOperation:
        return None


def get_int(row, idx):
    return int(get_value(row, idx))


class Parser():
    def __init__(self, f):
        self.book = open_workbook(file_contents=f.read())
        self.sheet = self.book.sheet_by_index(1)

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        row_index = sess = None
        try:
            sess = Session()
            bills = []
            title_row = self.sheet.row(10)
            issue_date = get_start_date(
                self.sheet.row(5), 2, self.book.datemode)
            if issue_date is None:
                raise BadRequest("Expected to find the issue date at cell C6.")

            for row_index in range(11, self.sheet.nrows):
                row = self.sheet.row(row_index)
                val = get_value(row, 1)
                if val is None or val == '':
                    break

                self._set_last_line(row_index, val)
                mpan_core = parse_mpan_core(str(get_int(row, 1)))
                comms = get_str(row, 2)

                settled_str = get_str(row, 3)
                if settled_str == 'Settled':
                    settlement_status = 'settlement'
                else:
                    settlement_status = 'non_settlement'

                start_date = get_start_date(row, 5, self.book.datemode)
                finish_date = get_finish_date(row, 6, self.book.datemode)

                meter_rate = get_dec(row, 7)
                net = round(get_dec(row, 8), 2)
                vat = round(get_dec(row, 9), 2)
                gross = round(get_dec(row, 10), 2)

                breakdown = {
                    'raw-lines': [str(title_row)], 'comms': comms,
                    'settlement-status': [settlement_status],
                    'meter-rate': [meter_rate], 'meter-gbp': net
                }

                bills.append(
                    {
                        'bill_type_code': 'N', 'kwh': Decimal(0), 'net': net,
                        'vat': vat, 'gross': gross, 'reads': [],
                        'breakdown': breakdown, 'account': mpan_core,
                        'issue_date': issue_date, 'start_date': start_date,
                        'finish_date': finish_date, 'mpan_core': mpan_core,
                        'reference': '_'.join(
                            (
                                start_date.strftime('%Y%m%d'),
                                finish_date.strftime('%Y%m%d'),
                                issue_date.strftime('%Y%m%d'),
                                mpan_core
                            )
                        )
                    }
                )
        except BadRequest as e:
            raise BadRequest(
                "Row number: " + str(row_index) + " " + e.description)
        finally:
            if sess is not None:
                sess.close()

        return bills
